<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * @ignore
 */
@include_once("DB.php");

/**
 * database API
 *
 * this class is a database abstraction api, that uses pear db
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class BufferedDbStream extends Object
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var  int    */   var $exists              = YANA_DATABASE_FAILURE;
    /** @var  Object */   var $database            = null;
    /** @var  string */   var $lastSelectedTable   = "";
    /** @var  string */   var $filename            = "";
    /** @var  array  */   var $queue               = array();
    /** @var  array  */   var $cache               = array();
    /** @var  array  */   var $joins               = array();
    /** @var  array  */   var $reservedSqlKeywords = null;
    /** @var  SML    */   var $lastModified        = null;

    /**#@-*/

   /**
    * @ignore
    * @access protected
    * @var object    */   var $structure           = null;


    /**
     * create a new instance
     *
     * Each database connection depends on a structure file
     * describing the database. Therefore you have to provide
     * a $filename of such structure file (without path or extension)
     *
     * These files are to be found in config/db/*.config
     *
     * @param string $filename
     * @param dbServer $server
     */
    function BufferedDbStream($filename = "", $server = null)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($filename); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($filename, "string");
        global $YANA;

        if (is_null($server)) {
            $server = new dbServer();
        }
        $this->database = $server->get();
        $this->filename = $filename;

        if (empty($this->database) || $this->isError($this->database)) {
            /* connection failed - abort */
            $this->exists = YANA_DATABASE_FAILURE;
            $this->handleError($this->database);
        } else {

            /* deactivate autocommit by default */
            $result = $this->database->autoCommit(false);
            if ($this->isError($result)) {
                $this->handleError($result);
            }

            if (empty($filename)) {
                /* no structure file available, estimate structure */
                $this->buildStructure(); /* ability to build structure on the fly */
                $this->exists = YANA_DATABASE_NO_FILE;
            } else {
                $this->structure = new StructureFile($this->filename);
                $test = $this->structure->init();
                if ($test === false) {
                    /* given file does not exist */
                    unset($this->structure);
                    $this->exists = YANA_DATABASE_NOT_AVAILABLE;
                    trigger_error("Unable to initialize database. Structure file '${filename}' not found.", E_USER_WARNING);
                } else {
                    /* all fine, proceed */
                    $this->exists = YANA_DATABASE_OK;
                }
            }
        }
    }

    /**
     * Commit current transaction
     *
     * This writes all changes to the database
     *
     * @access public
     * @return boolean
     */
    function write()
    {
        if (!$this->_check_writeable()) {
            return false;
        }

        if (count($this->queue) == 0) {
            /* Buffer empty */
            return true;
        } else {
            for ($i=0; $i<count($this->queue); $i++)
            {
                /* SQL statement */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][0])) {
                    $sqlStmt = trim($this->queue[$i][0]);
                } else if (is_string($this->queue[$i])) {
                    $sqlStmt = trim($this->queue[$i]);
                } else {
                    continue;
                }
                /* trigger */
                if (is_array($this->queue[$i]) && isset($this->queue[$i][1]) && is_array($this->queue[$i][1])) {
                    $args = $this->queue[$i][1];
                } else {
                    $args = null;
                }
                /* ignore empty queue elements */
                if (empty($sqlStmt)) {
                    continue;
                /* for security reasons all delete queries will automatically be limited to 1 row at a time */
                } else if (preg_match('/^DELETE /', $sqlStmt)) {
                    $result = $this->database->limitQuery($sqlStmt,0,1);
                } else {
                    $result = $this->database->query($sqlStmt);
                }
                assert('!isset($success); // cannot redeclare variable $success');
                if ($this->isError($result)) {
                    /* fire trigger */
                    if (!is_null($args)) {
                        $success = false;
                        $this->structure->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                    /* rollback on error */
                    $this->handleError($result,'SQL ERROR: Statement [code]'.$sqlStmt.'[/code]');
                    $result = $this->database->rollback();
                    if ($this->isError($result)) {
                          $this->handleError($result, "SQL ERROR: Unable to rollback changes. Database might contain corrupt data. Please inform webmaster about time of event and anything you might have done to cause this failure.");
                    }
                    return false;
                } else {
                    if (!is_null($args)) {
                        $success = true;
                        $this->structure->trigger(YANA_TRIGGER_AFTER, $args[0], $args[1], $args[2], $success);
                    }
                }
                unset($success);
            } /* end foreach */
            $result = $this->database->commit();
            if ($this->isError($result)) {
                if (!is_null($this->lastModified)) {
                    $this->lastModified->reset();
                }
                $this->handleError($result);
                return false;
            } else {
                if (!is_null($this->lastModified)) {
                    $this->lastModified->write();
                }
                $this->queue = array();
                return true;
            }
        }
    }

    /**
     * get values from the database
     *
     * This returns the values at adress $key
     * starting from $offset and limited to
     * $limit results.
     *
     * The $key parameter has two synopsis.
     * <ul>
     *     <li>
     *         If $key is a string, this parameter is interpreted
     *         as the address of the values you want to retrieve.
     *     </li>
     *     <li>
     *         If $key is an associative array, it overwrites any of the
     *         paramters you provide, without the need to enter default
     *         parameters - this might come in handy in certain situations.
     *     </li>
     * </ul>
     *
     * The array $key can have the following values
     * <ul>
     *     <li>  $key      = the address of the value(s) to retrieve (mandatory) </li>
     *     <li>  $where    = a search string (optional) </li>
     *     <li>  $search   = alias of $where (deprecated) </li>
     *     <li>  $order_by = array of columns to order the resultset by (optional) </li>
     *     <li>  $desc     = boolean, if true results will be ordered in descending, otherwise in ascending order (optional) </li>
     *     <li>  $offset   = integer, the number of the first result to be returned (optional) </li>
     *     <li>  $limit    = integer, maximum number of results to return (optional) </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $order_by has two synopsis.
     * <ul>
     *     <li>
     *         $order_by is the name of the column in the current table to order the resultset by.
     *     </li>
     *     <li>
     *         $order_by is a numeric array of strings, where each element
     *         is the name of a column in the current table.
     *         The resultset will get ordered by the values of these columns
     *         in the direction in which they are provided.
     *         This feature became available in version 2.8.5
     *     </li>
     * </ul>
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * Note: The $desc parameter became available in version 2.8.5
     *
     * @access public
     * @param  array|string $key        the address of the value(s) to retrieve
     * @param  array|string $where      a search string
     * @param  array        $order_by   a list of columns to order the resultset by
     * @param  int          $offset     the number of the first result to be returned
     * @param  int          $limit      maximum number of results to return
     * @param  bool         $desc       if true results will be ordered in descending, otherwise in ascending order
     * @return mixed
     */
    function get($key, $where = array(), $order_by = array(), $offset = 0, $limit = 0, $desc = false)
    {
        /* 1.1 handle $key array */
        if (is_array($key)) {
            if (isset($key['where'])) {
                $where = $key['where'];
            } else if (isset($key['search'])) {
                $where = $key['search'];
            }
            if (isset($key['order_by'])) {
                $order_by = $key['order_by'];
            }
            if (isset($key['desc'])) {
                $desc = $key['desc'];
            }
            if (isset($key['offset'])) {
                $offset = $key['offset'];
            }
            if (isset($key['limit'])) {
                $limit = $key['limit'];
            }
            if (isset($key['key'])) {
                $key = $key['key'];
            } else {
                return false;
            }
        }
        /* 1.2 check input */
        /* 1.2.1 first argument */
        if (is_string($key)) {
            $key = strtolower($key);
            $table = explode('.', $key);
            $table = $table[0];
        } else {
            trigger_error("Wrong argument type for argument 1. String expected found '".gettype($key)."' instead.", E_USER_WARNING);
            return false;
        }
        /* 1.2.2 next argument */
        if (!is_string($where) && !is_array($where)) {
            trigger_error("Wrong argument type for argument 2. String or array expected found '".gettype($where)."' instead.", E_USER_WARNING);
            return false;
        }
        /* 1.2.3 next argument */
        if (is_string($order_by)) {
            if (!empty($order_by)) {
                $order_by = array($order_by);
            } else {
                $order_by = array();
            }
        }
        if (is_array($order_by)) {
            if (count($order_by) < 1) {
                $order_by = "";
            } else {
                /* for strict structures also test if input is a column */
                if ($this->structure->isStrict()) {
                    assert('!isset($column);');
                    assert('!isset($m);');
                    foreach ($order_by as $column)
                    {
                        if (empty($column)) {
                            continue;
                        }
                        /* if a table name is provided */
                        if (is_string($column) && strpos($column,'.') !== false) {
                            preg_match('/^(.+)\.(.*)$/U',$column,$m);
                            if (!$this->structure->isColumn($m[1],$m[2])) {
                                trigger_error("The 'order_by' parameter is invalid. There is no column '".$m[2]."' in table '".$m[1]."'.",E_USER_WARNING);
                                return false;
                            }
                        /* if no table name is provided, use current table */
                        } else if (!$this->structure->isColumn($table,$column)) {
                            trigger_error("The 'order_by' parameter is invalid. There is no column '${column}' in table '${table}'.",E_USER_WARNING);
                            return false;
                        }
                    } /* end foreach */
                    unset($m, $column);
                } /* end if */
                /* quote ids */
                $order_by = array_map(array($this,'quoteId'),$order_by);
                /* convert array into a string */
                $order_by = 'order by '.implode(',',$order_by);
            } /* end if */
        } else {
            trigger_error("Wrong argument type for argument 3. String or Array expected found '".gettype($order_by)."' instead.", E_USER_WARNING);
            return false;
        }
        $order_by = strtolower($order_by);
        /* 1.2.4 next argument */
        if (!is_int($offset)) {
            trigger_error("Wrong argument type for argument 4. Integer expected found '".gettype($offset)."' instead.", E_USER_WARNING);
            return false;
        } else {
            assert('$offset >= 0; /* SQL select statement: the offset '.$offset.' is out of range. */');
        }
        /* 1.2.5 next argument */
        if (!is_int($limit)) {
            trigger_error("Wrong argument type for argument 5. Integer expected found '".gettype($limit)."' instead.", E_USER_WARNING);
            return false;
        } else {
            assert('$limit >= 0; /* SQL select statement: the limit '.$limit.' is out of range. */');
        }
        /* 1.2.6 last argument */
        if (!is_bool($desc)) {
            trigger_error("Wrong argument type for argument 6. Boolean expected found '".gettype($desc)."' instead.", E_USER_WARNING);
            return false;
        } else {
            /* have values in descending order if $desc is set to true */
            if ($desc === true) {
                $order_by = trim($order_by.' desc');
            }
        }

        /* 1.3 check availability of database connection */
        if (!$this->_check_available()) {
            return false;
        }

        if ($offset<0) {
            $offset = 0;
        }
        if ($limit<0) {
            $limit = 0;
        }
        /* 2 handle request */
        $result = $this->_computeRequest($key, $where, NULL, $offset, $limit, $order_by);

        /* 3 return result */
        return $result;
    }

    /**
     * update or insert row
     *
     * insert $value at position $key
     *
     * If $key already exists, the previous value
     * gets updated, else the value is created
     *
     * This function returns bool(true) on success
     * and bool(false) on error. Note, that this
     * function does not auto-commit. This means,
     * changes to the database will NOT be saved
     * unless you call $BufferedDbStream->write().
     *
     * @param string $key
     * @param mixed $value
     * @access public
     * @return boolean
     */
    function insert($key, $value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($key, "string");

        $key = strtolower($key);

        if (!$this->_check_writeable()) {
            return false;
        }

        /* check whether the row has been modified since last access */
        if ($this->structure->isStrict()) {
            if (isset($_SESSION['transaction_isolation_created'])) {
                if ($this->_lastModified($key) > $_SESSION['transaction_isolation_created']) {
                    $this->handleError("SQL ERROR: Input data has been changed since last access. Updating this row would overwrite existing changes. Insert/Update aborted for security reasons.");
                    return false;
                }
            }
        }

        if ($this->_computeRequest($key, "", $value)) {
            return Hashtable::set($this->cache, $key, $value);
        } else {
            $this->handleError("SQL ERROR: Insert on database failed for key '${key}'.", preg_replace('/\s/', ' ', SML::encode($value)));
            return false;
        }

    }

    /**
     * remove one row
     *
     * For security reasons all delete queries will automatically be limited to 1 row at a time.
     * While this might be seen as a limitation the far more valueable advantage is,
     * no user is able to destroy a whole table - wether by intention or by accident -
     * in only one query. (At least not via this API.)
     *
     * The function returns bool(true) on success and bool(false) on error.
     *
     * Since version 2.8.5 the parameter $where has two synopsis.
     * <ul>
     *     <li>
     *         $where is a comma seperated list of column-value pairs using the following syntax:
     *         <pre>column1=value1,column2=value2,...,columnN=valueN</pre>
     *         Note that comparison takes place using the SQL operator 'LIKE', except for
     *         columns that are keys (either a primary key or a foreign key), which are compared
     *         using the SQL operator '='.
     *     </li>
     *     <li>
     *         $where is a two-dimensional numeric array and its entries follow this syntax:
     *         <pre>array(array(0=>column1,1=>value1,2=>operator1),...)</pre>
     *         The "column" name is mandatory and needs to be a string.
     *         The "value" is mandatory and needs to be a scalar value.
     *         The "operator" is optional and defaults to 'LIKE'.
     *         It needs to be a string and can be one of the following:
     *         <pre> '=', 'LIKE', '<', '>', '!=', '<=', '>=', '==' (alias of '=') </pre>
     *     </li>
     * </ul>
     * The second alternative offers more flexibility and should be prefered over the first one.
     *
     * @access  public
     * @param   string        $key   the address of the row that should be removed
     * @param   string|array  $where a search string
     * @return  bool
     */
    function remove($key, $in_where = array())
    {
        /* 1 check input */
        if (!is_string($key)) {
            trigger_error("Wrong argument type for argument 1. String expected found '".gettype($key)."' instead.", E_USER_WARNING);
            return false;
        }
        if (!is_string($in_where) && !is_array($in_where)) {
            trigger_error("Wrong argument type for argument 2. String or array expected found '".gettype($in_where)."' instead.", E_USER_WARNING);
            return false;
        }

        /* untaint input */
        $key = strtolower($key);

        if (!$this->_check_writeable()) {
            return false;
        }

        $keyArray = explode(".", $key);
        settype($keyArray, "array");
        if (!$this->structure->isTable($keyArray[0])) {
            trigger_error("Argument 1 is invalid, '".$keyArray[0]."' is not a table name.", E_USER_WARNING);
            return false;
        }
        $this->lastSelectedTable = $keyArray[0];

        if ((is_string($in_where) && !empty($in_where)) || (is_array($in_where) && count($in_where) > 0)) {
            $where = $this->_computeSearchString($keyArray[0], $in_where, count($keyArray) == 1);
            if ($where === false) {
                trigger_error("Argument 2 is invalid. Where available, see log file for details.", E_USER_WARNING);
                return false;
            }
        } else {
            $where = "";
        }

        if ((count($keyArray) < 2 && empty($where)) || (count($keyArray) > 2) || ($keyArray[1] == '0') || (@$this->structure->constraint('DELETE', $keyArray[0], $keyArray[1]) === false)) {
            trigger_error("Argument 1 is invalid. Must be a valid address.", E_USER_WARNING);
            return false;
        }

        /* check constraint */
        if ($this->structure->constraint('DELETE', $keyArray[0]) === false) {
            return false;
        }

        /* fire trigger */
        assert('!isset($value); // cannot redeclare variable $value');
        $value = null;
        @$this->structure->trigger(YANA_TRIGGER_BEFORE, 'DELETE', $keyArray[0], null, $value);
        unset($value);

        /* create SQL */
        $sql = array();
        if ($keyArray[1] == "*") {
            /* remove last entry */
            $sql = array();
            $sql[0] = "DELETE FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0]).( (!empty($where)) ? " WHERE $where " : '' )." order by ".$this->structure->getPrimaryKey($keyArray[0])." desc";
            $sql[1] = array('DELETE',$keyArray[0],null,null);
        } else {
            /* remove entry with id $keyArray[1] */
            if (isset($keyArray[1]) && $keyArray[1] != '') {
                if (!empty($where)) {
                    $where .= ' AND ';
                }
                $where .= $this->structure->getPrimaryKey($keyArray[0])." = ".$this->quote($keyArray[1]);
            }
            $sql = array();
            $sql[0] = "DELETE FROM ".$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0])." WHERE ".( ($where) ? " $where " : '' );
            $sql[1] = array(0=>'DELETE',1=>$keyArray[0],2=>null);
        }
        array_push($this->queue,$sql);

        /* backup entry before deleting it */
        global $YANA;
        if (isset($YANA)) {
            assert('!isset($result); // cannot redeclare variable $result');
            $result = $this->get($key,$in_where,'',0,1);
            if (!empty($result)) {
                if (is_string($result)) {
                    $YANA->report(new log('Deleting entry from '.$key.':', preg_replace('/\s/', ' ', $result)));
                } else {
                    $YANA->report(new log('Deleting entry from '.$key.':', $result));
                }
            }
        }

        return true;
    }

    /**
     * join the resultsets for two tables
     *
     * results in an INNER JOIN $table1, $table2 WHERE $table1.$key1 = $table2.$key2
     *
     * @access public
     * @param string $table1
     * @param string $table2 (optional)
     * @param string $key1 (optional)
     * @param string $key2 (optional)
     * @return boolean
     */
    function join($table1, $table2 = "", $key1 = "", $key2 = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table1); // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_string($table2); // wrong argument type for argument 2, string expected');
        /* argument 3 */ assert('is_string($key1);   // wrong argument type for argument 3, string expected');
        /* argument 4 */ assert('is_string($key2);   // wrong argument type for argument 4, string expected');

        if (!is_string($table1) || !preg_match("/^[\w\d-_]*$|^\*$/si", $table1)) {
            trigger_error("Argument 1 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid table name.", E_USER_WARNING);
            return false;
        }
        if (!is_string($table2) || !preg_match("/^[\w\d-_]*$|^\*$/si", $table2)) {
            trigger_error("Argument 2 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid table name.", E_USER_WARNING);
            return false;
        }
        if (!is_string($key1) || !preg_match("/^[\w\d-_]*$|^\*$/si", $key1)) {
            trigger_error("Argument 3 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid column name.", E_USER_WARNING);
            return false;
        }
        if (!is_string($key2) || !preg_match("/^[\w\d-_]*$|^\*$/si", $key2)) {
            trigger_error("Argument 4 is invalid for function ".get_class($this)."->".__FUNCTION__."(). Must be a valid column name.", E_USER_WARNING);
            return false;
        }

        /* lower-case input */
        $table1 = strtolower($table1);
        $table2 = strtolower($table2);
        $key1   = strtolower($key1);
        $key2   = strtolower($key2);

        if (!$this->_check_available()) {
            return false;
        }

        if ($table1 == "" || $table1 == "*") {
            if (!$this->lastSelectedTable) {
                trigger_error("Wrong parameter count for function ".get_class($this)."->".__FUNCTION__."(). Unable to join tables, since no table has been selected for function join.", E_USER_WARNING);
                return false;
            } else {
                $table1 = $this->lastSelectedTable;
            }
        }

        if ($this->structure->isStrict()) {

            if (!$this->structure->isTable($table1) || (!empty($table2) && !$this->structure->isTable($table2))) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Table does not exist in structure file.");
                return false;
            }
            if (!empty($key1) && !$this->structure->isColumn($table1, $key1)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key1}' does not exist in table '${table1}'.");
                return false;
            }
            if (!empty($table2) && !empty($key2) && !$this->structure->isColumn($table2, $key2)) {
                $this->handleError("Cannot join tables '${table1}' and '${table2}'. Field '${key2}' does not exist in table '${table2}'.");
                return false;
            }
        }

        if (!empty($table2)) {
            $this->joins[$table1][$table2]['TABLE'] = $table2;
            if (!empty($key1)) {
                $this->joins[$table1][$table2]['FKEY']  = $key1;
            }
            if (!empty($key1) && !empty($key2)) {
                $this->joins[$table1][$table2]['TKEY']  = $key2;
            }
            return true;
        } else {
            $this->joins[$table1] = array();
            return true;
        }

    }

    /**
     * optional API bypass
     *
     * Send a sql-statement directly to the PEAR database API, bypassing this API.
     *
     * Note: for security reasons this only sends one single SQL statement at a time.
     * This is done by checking the input for a semicolon followed by anything but
     * whitespace. If such input is found, an E_USER_WARNING is issued and the 
     * function will return bool(false).
     *
     * While bypassing the API leaves nearly all of the input checking to you, this
     * is meant to prevent at least a minimum of the common SQL injection attacks.
     * A known attack is to try to terminate a current statement with ';' and afterwards
     * "inject" their own stuff as a second statement. The common attack vector usually
     * is unchecked form data.
     *
     * If you want to send a sequence of statements, call this function multiple times.
     *
     * The function will return bool(false) if the database connection or the
     * PEAR API is not available and otherwise will whatever PEAR sends back as the
     * result of your statement.
     *
     * Note: when database usage is disabled via the administrator's menu,
     * the PEAR-DB API can not be used and this function will return bool(false).
     *
     * @access public
     * @param  string $sqlStmt the SQL statement to execute
     * @return mixed
     */
    function query($sqlStmt)
    {
        /* 1 check input type */
        if (!is_string($sqlStmt)) {
            trigger_error("Wrong argument type for argument 1. String expected found '".gettype($sqlStmt)."' instead.", E_USER_WARNING);
            return false;
        /* 2 check input syntax */
        } else if (is_int(strpos($sqlStmt, ';')) && preg_match("/;.*\S.*$/s", $sqlStmt)) {
            trigger_error("A semicolon has been found in the current input '${sqlStmt}', indicating multiple queries.\n\t\tAs this might be the result of a hacking attempt it is prohibited for security reasons and the queries won't be executed.", E_USER_WARNING);
            return false;
        /* 3 check if database connection is available */
        } else if (!$this->_check_available()) {
            return false;
        /* 4 check if database connection is available */
        } else {
            return $this->database->query($sqlStmt);
        }
    }

    /**
     * query building
     *
     * This function builds queries from input data.
     *
     * @access  private
     * @param   string        $key
     * @param   string|array  $search (optional)
     * @param   mixed         $value  (optional)
     * @param   int           $offset (optional)
     * @param   int           $limit  (optional)
     * @param   string        $order_by  (optional)
     * @return  mixed
     *
     * @ignore
     */
    function _computeRequest($key, $search = array(), $value = NULL, $offset = 0, $limit = 0, $order_by = "")
    {
        assert('is_string($key) && !empty($key);');
        assert('is_string($search) || is_array($search);');
        assert('is_int($offset) && $offset >= 0;  /* SQL select statement: the offset '.$offset.' is out of range. */');
        assert('is_int($limit) && $limit >= 0;    /* SQL select statement: the limit '.$limit  .' is out of range. */');
        assert('is_string($order_by);');

        settype($offset, "integer");
        settype($limit, "integer");

        if (!$this->_check_available()) {
            return false;
        }

        $key = strtolower($key);
        $where = "";

        /* int */ $returnedType = 0; /* 0 = unknown, 1 = table, 2 = row, 3 = field, 4 = column */
        /* int */ $queryType    = 0; /* 0 = unknown, 1 = select, 2 = update, 3 = insert */
        /* int */ $useWildcard  = 0; /* 0 = none, 1 = foo.*, 2 = foo.*.foo */
        /* string */ $arrayKey  = '';

        /* 1) check input */

        /* 1.1 check if search is valid */
        if (!is_array($search)) {
            settype($search, "string");
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\@\. \&;])+)*$/i", $search)) {
                trigger_error("Syntax error. The input '${search}' is not a valid search string.", E_USER_WARNING);
                return false;
            }
        }
        /* 1.2 check if key is valid */
        settype($key, "string");
        $key = strtolower($key);
        if (!preg_match("/^[\w\d-_]+(\.([\w\d-_]*|\*|\?)){0,}(\.\*)?$/i", $key)) {
            trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
            return false;
        }

        /* 2) prepare input for computation */

        if (preg_match("/\.\*?$/", $key)) {
            $useWildcard = 1;
        } else if (preg_match("/[\w\d-_]+\.\?\.([\w\d-_]+)$/i", $key)) {
            $key = preg_replace("/([\w\d-_]+)\.\?\.([\w\d-_]+)$/i", '${1}.*.${2}', $key);
            $useWildcard = 2;
        } else if (preg_match("/[\w\d-_]+\.\*\.([\w\d-_]+)$/i", $key)) {
            $useWildcard = 2;
            $returnedType = 4;
        } else {
            $useWildcard = 0;
        }

        $key = preg_replace("/\.(\*)?$/", '', $key);
        $keyArray = explode(".", $key);

        /* 3) resolve foreign keys to get true adress */
        if (count($keyArray) > 3) {
            if ($this->structure->getType($keyArray[0],$keyArray[2]) == 'array') {
                $arrayKey = implode('.', array_slice($keyArray, 3));
                $keyArray = array_slice($keyArray, 0,3);
            } else {
                while (count($keyArray) > 3 && $this->structure->isForeignKey($keyArray[0],$keyArray[2]))
                {
                    assert('!isset($a); /* cannot redeclare variable $a */');
                    $a = $this->_computeRequest($keyArray[0].".".$keyArray[1].".".$keyArray[2]);
                    if (!$a) {
                        $this->handleError("SQL ERROR: Unable to resolve foreign key '${key}'. This may mean one of the associated tables contains inconsistent data. Check if foreign key association has been broken.");
                        return false;
                    }
                    assert('!isset($foreignKeys); /* cannot redeclare variable $foreignKeys */');
                    $foreignKeys = $this->structure->getForeignKeys($keyArray[0]);
                    assert('!isset($foreignTable); /* cannot redeclare variable $foreignTable */');
                    $foreignTable = $foreignKeys[strtoupper($keyArray[2])];
                    assert('!isset($resultType); /* cannot redeclare variable $resultType */');
                    $resultType = $this->structure->getType($foreignTable,$keyArray[2]);
                    if ($resultType) {
                        @settype($a, $resultType);
                    }
                    @ array_shift($keyArray); @ array_shift($keyArray); @ array_shift($keyArray);
                    array_unshift($keyArray, $a);
                    array_unshift($keyArray, $foreignTable);
                    unset($a, $foreignKeys, $foreignTable, $resultType);
                }
            }
        } else {
            /* 3.1) inheritance */
            assert('!isset($primaryKey);');
            $primaryKey = strtoupper($this->structure->getPrimaryKey($keyArray[0]));
            if ($this->structure->isForeignKey($keyArray[0],$primaryKey)) {

                $foreignKeys = $this->structure->getForeignKeys($keyArray[0]);
                $foreignTable = $foreignKeys[$primaryKey];
                $this->join($keyArray[0],$foreignTable,$primaryKey,$primaryKey);
            }
        }

        if(count($keyArray) > 3) {
            trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
            return false;
        } else {
            $key = implode('.', $keyArray);
            if ($returnedType == 0) {
                $returnedType = count($keyArray);
            }
        }

        /* 4) now that there is a resolved adress: if $value is set, check first if it is valid before going on */

        if ($returnedType>1) {
            /* Let's check the primary key first */
            $primaryKey = $this->structure->getPrimaryKey($keyArray[0]);
            $keyArray[1] = $this->structure->untaintInput($keyArray[0], $primaryKey, $keyArray[1], YANA_ESCAPE_SLASHED);
        }

        if (!is_null($value)) {

            if (!$this->_check_writeable()) {
                return false;
            }

            if ($returnedType==1 && ($useWildcard==1)) {
                $returnedType = 2;
                $queryType    = 3;
            }
            if ($returnedType==4) {
                $returnedType = count($keyArray);
            }

            if (!$queryType) {
                if ($returnedType==3) {
                    /* Inserting/Updating a field */
                    $queryType = ($this->exists($keyArray[0].'.'.$keyArray[1])) ? 2 : 3;
                } else {
                    /* Inserting/Updating smt. else */
                    $queryType = ($this->exists($key)) ? 2 : 3;
                }
            }

            switch ($returnedType)
            {

                /* Inserting/Updating a table */
                case 1:
                    if ($this->structure->isStrict()) {
                        $this->handleError("Inserting/Updating a table is illegal in strict mode. Operation aborted.");
                    }
                    return false;
                break;

                /* Updating a row */
                case 2:
                    settype($value, "array");
                    $value = array_change_key_case($value, CASE_LOWER);

                    /* untaint raw data */
                    assert('!isset($a); /* cannot redeclare variable $a */');
                    assert('!isset($b); /* cannot redeclare variable $b */');
                    foreach ($value as $a => $b)
                    {
                        if ($queryType == 2 && $this->structure->isReadonly($keyArray[0],$a)) {
                            $this->handleError("SQL ERROR: Database is readonly. Insert/Update operation aborted.");
                            return false;
                        }
                        $type = $this->structure->getType($keyArray[0],$a);
                        switch ($type)
                        {
                            case 'text':
                                $escape = YANA_ESCAPE_USERTEXT;
                            break;
                            case 'string':
                                $escape = YANA_ESCAPE_LINEBREAK;
                            break;
                            default:
                                $escape = YANA_ESCAPE_NONE;
                            break;
                        } /* end switch */
                        $value[$a] = $this->structure->untaintInput($keyArray[0], $a, $b, $escape);
                        /* if it is a foreign key, check wether there is a field with that id */
                        $foreignTable = $this->structure->getTableByForeignKey($keyArray[0],$a);
                        if ($this->structure->isForeignKey($keyArray[0],$a) && !$this->exists("$foreignTable.$b")) {
                            /* OK - there is none. Now let's check if there is one recently inserted field in cache */
                            if (!isset($this->cache[$foreignTable][$b])) {
                                $this->handleError("SQL Error: The value '${b}' for attribute '${a}' refers to a non-existing entry in table '${foreignTable}'. Insert/Update operation aborted.");
                                return false;
                            }
                        }
                        unset($foreignTable);
                    } /* end foreach */
                    unset($a,$b);

                    /* check if all required data is available */
                    if ($queryType == 3) {
                        if ($this->structure->checkRow($keyArray[0],$value) === false) {
                            return false;
                        }
                    } else if ($queryType == 2) {
                        if ($this->structure->checkRow($keyArray[0],$value,false) === false) {
                            return false;
                        }
                    } /* end if */
                break;

                /* Updating a field */
                case 3:

                    /* untaint raw data */
                    if ($queryType == 2 && $this->structure->isReadonly($keyArray[0],$keyArray[2])) {
                        $this->handleError("SQL ERROR: Database is readonly. Insert/Update operation aborted.");
                        return false;
                    }
                    if ($this->structure->getType($keyArray[0], $keyArray[2]) == 'string') {
                        $escape = YANA_ESCAPE_LINEBREAK;
                    } else {
                        $escape = YANA_ESCAPE_NONE;
                    }
                    $value = $this->structure->untaintInput($keyArray[0], $keyArray[2], $value, $escape);

                    /* if it is a foreign key, check wether there is a field with that id */
                    if ($this->structure->isForeignKey($keyArray[0], $keyArray[2])) {
                        $foreignTable = $this->structure->getTableByForeignKey($keyArray[0],$keyArray[2]);
                        if (!$this->exists("${foreignTable}.${value}")) {
                            $this->handleError("SQL Error: The value '${value}' for attribute '".$keyArray[2]."' refers to a non-existing entry in table '${foreignTable}'. Insert/Update operation aborted.");
                            return false;
                        }
                    }

                break;

                /* error */
                default:
                    trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                    return false;
                break;

            }

        } else { /* $value is not set - since there no value to set this will be a simple read operation */
            $queryType = 1;
        }

        $this->lastSelectedTable = $keyArray[0];

        /* 5) build sql-statement */

        switch ($queryType)
        {
            case 1:
                $sqlStmt = "SELECT";
                if (empty($keyArray[2])) {
                    $sqlStmt .= " * ";
                } else {
                    $sqlStmt .= " ".$keyArray[2]." ";
                }
                $sqlStmt .= "FROM ".$this->quoteId($keyArray[0])." ";
            break;

            case 2:
                if ($returnedType == 2) {
                    if (!is_array($value)) {
                        trigger_error("Syntax error. The argument \$value '${value}' is not valid.", E_USER_WARNING);
                        return false;
                    }
                    $sqlStmt = $this->_buildUpdate($keyArray[0],$value);
                } else {
                    $sqlStmt = $this->_buildUpdate($keyArray[0],array($keyArray[2] => $value));
                }
            break;

            case 3:
                if ($returnedType == 2) {
                    $temp = $this->structure->getPrimaryKey($keyArray[0]);
                    if (!isset($value[$temp]) && !empty($keyArray[1])) {
                        $value[$temp] = $keyArray[1];
                    }
                    $sqlStmt = 'INSERT INTO '.$this->quoteId(YANA_DATABASE_PREFIX.$keyArray[0]).' '.$this->_sqlValues($value).' ';
                    unset($temp);
                } else {
                    $this->handleError("SQL ERROR: Invalid insert operation. You may only insert rows (not fields or tables).");
                    return false;
                }
            break;

            default:
                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                return false;
            break;

        }

        /* 6) resolve joins */

        if ($queryType < 3 && $queryType > 0 && $useWildcard < 2) {
            $where = (!empty($keyArray[1])) ? $this->structure->getPrimaryKey($keyArray[0])." = ".$this->quote(addslashes($keyArray[1]))." " : "";
        }

        if ($queryType == 1) {
            assert('!isset($JOIN); /* cannot redeclare variable $JOIN */');
            $JOIN  = array();
            if (isset($this->joins[$keyArray[0]])) {
                $JOIN = $this->joins[$keyArray[0]];
            }
            settype($JOIN, "array");

            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($JOIN as $b)
            {
                if (count($JOIN) == 1 && !$b['TKEY'] && !$b['FKEY']) {
                    $sqlStmt .= "NATURAL JOIN ".$this->quoteId(YANA_DATABASE_PREFIX.$b['TABLE'])." ";
                } else {
                    $sqlStmt .= ", ".$this->quoteId(YANA_DATABASE_PREFIX.$b['TABLE'])." ";
                    if ($b['FKEY'] && $b['TKEY']) {
                          $where .= ( (!empty($where)) ? "AND " : "" ).YANA_DATABASE_PREFIX.$keyArray[0].'.'.$b['FKEY'].' = '.YANA_DATABASE_PREFIX.$b['TABLE'].'.'.$b['TKEY'].' ';
                    } else if ($b['FKEY']) {
                          $where .= ( (!empty($where)) ? "AND " : "" ).YANA_DATABASE_PREFIX.$keyArray[0].'.'.$b['FKEY'].' = '.YANA_DATABASE_PREFIX.$b['TABLE'].'.'.$b['FKEY'].' ';
                    } else {
                          /* intenionally left blank */
                    }
                }
            } /* end foreach */
            unset($b);
        }

        /* 7) resolve search clause */
        $where = trim($where);
        if ((is_string($search) && !empty($search)) || (is_array($search) && count($search) > 0)) {
            if (!empty($where)) {
                $where .= ' AND ';
            }
            if (count($keyArray) == 1 || $keyArray[1] == '?' || $keyArray[1] == '*') {
                $where .= $this->_computeSearchString($keyArray[0], $search, true);
            } else {
                $where .= $this->_computeSearchString($keyArray[0], $search, false);
            }
        }
        $sqlStmt .= ( (!empty($where)) ? "WHERE $where " : " " ).$order_by;

        /* 8) check constraints and triggers */
        assert('!isset($triggerArgs);');
        /* array */ $triggerArgs = null; /* 0 => OPERATION, 1 => TABLE, 2 => COLUM */
        switch ($queryType) /* 0 = unknown, 1 = select, 2 = update, 3 = insert */
        {
            case 1:
                /* intentionally left blank */
            break;
            case 2:
                if (@$this->structure->constraint('UPDATE', $keyArray[0], $keyArray[2], $value) === false) {
                    return false;
                } else {
                    @$this->structure->trigger(YANA_TRIGGER_BEFORE, 'UPDATE', $keyArray[0], $keyArray[2], $value);
                }
                $triggerArgs = array('UPDATE',$keyArray[0],@$keyArray[2]);
            break;
            case 3:
                if (@$this->structure->constraint('INSERT', $keyArray[0], $keyArray[2], $value) === false) {
                    return false;
                } else {
                    @$this->structure->trigger(YANA_TRIGGER_BEFORE, 'INSERT', $keyArray[0], $keyArray[2], $value);
                }
                $triggerArgs = array('INSERT',$keyArray[0],@$keyArray[2]);
            break;
            default:
                trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                return false;
            break;
        }

        if ($queryType == 2 || $queryType == 3) {
            if (is_null($triggerArgs)) {
                array_push($this->queue,$sqlStmt);
            } else {
                array_push($this->queue,array(0=>$sqlStmt,1=>$triggerArgs));
            }
            return true;
        } else if ($queryType == 1) {
            if ($offset>0 || $limit>0) {
                $result = $this->database->limitQuery($sqlStmt, $offset, $limit);
            } else {
                $result = $this->database->query($sqlStmt);
            }
            if ($this->isError($result)) {
                $this->handleError($result, "SQL WARNING: Query '${sqlStmt}' on database failed");
            } else {
                $REF = array();
                $id  = $this->structure->getPrimaryKey($this->getTable());

                for ($i=0; $i<$result->numRows(); $i++)
                {
                    $row = $result->fetchRow(DB_FETCHMODE_ASSOC,$i);
                    if (!is_array($row)) {
                        $this->handleError("DB NOTICE: The returned data for statement '${sqlStmt}' must be of type array. Instead database returned the following value '${row}'. The result was considered to be an error.");
                        break;
                    }
                    assert('!isset($a); /* cannot redeclare variable $a */');
                    assert('!isset($b); /* cannot redeclare variable $b */');
                    foreach ($row as $a => $b)
                    {
                        if (!is_null($b)) {
                            switch ($returnedType)
                            {
                                case 1:
                                    if (@$this->structure->constraint('SELECT', $keyArray[0], null, $b) !== false) {
                                        if ($this->structure->getType($keyArray[0],$a) !== 'array') {
                                            $REF[strtoupper($row[$id])][strtoupper($a)] = stripslashes($b);
                                        } else {
                                            if (is_array($b)) {
                                                $REF[strtoupper($row[$id])][strtoupper($a)] = $b;
                                            } else {
                                                $REF[strtoupper($row[$id])][strtoupper($a)] = SML::getFile(explode("\n",$b), CASE_UPPER);
                                            }
                                        }
                                    } else {
                                        unset($REF[strtoupper($row[$id])]);
                                        break;
                                    }
                                break;

                                case 2:
                                    if (@$this->structure->constraint('SELECT', $keyArray[0], null, $b) !== false) {
                                        if ($this->structure->getType($keyArray[0],$a) !== 'array') {
                                            $REF[strtoupper($a)] = stripslashes($b);
                                        } else {
                                            if (is_array($b)) {
                                                $REF[strtoupper($a)] = $b;
                                            } else {
                                                $REF[strtoupper($a)] = SML::getFile(explode("\n",$b), CASE_UPPER);
                                            }
                                        }
                                    } else {
                                        return array();
                                    }
                                break;

                                case 3:
                                    if (@$this->structure->constraint('SELECT', $keyArray[0], $keyArray[2], $b) !== false) {
                                        if ($this->structure->getType($keyArray[0],$keyArray[2]) !== 'array') {
                                            $REF = $b;
                                        } else {
                                            if (is_array($b)) {
                                                $REF = $b;
                                            } else {
                                                $REF = SML::getFile(explode("\n",$b), CASE_UPPER);
                                            }
                                            if ($arrayKey !== "") {
                                                $REF = Hashtable::get($REF, strtolower($arrayKey));
                                                if (is_null($REF)) {
                                                    $REF = false;
                                                }
                                            }
                                        }
                                    } else {
                                        return false;
                                    }
                                break;

                                case 4:
                                    if ($this->structure->constraint('SELECT', $keyArray[0], $keyArray[2], $b) !== false) {
                                        if ($this->structure->getType($keyArray[0],$keyArray[2]) !== 'array') {
                                            $REF[$i] = stripslashes($b);
                                        } else {
                                            if (is_array($b)) {
                                                $REF[$i] = $b;
                                            } else {
                                                $REF[$i] = SML::getFile(explode("\n",$b), CASE_UPPER);
                                            }
                                        }
                                    }
                                break;

                                default:
                                    trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
                                    return false;
                                break;
                            } /* end switch */
                        } /* end if */
                    } /* end foreach */
                    unset($a,$b);
                } /* end for */
                return $REF;
            } /* end else */
            return false;
        } else { /* end else if ($queryType == 1) */
            trigger_error("Syntax error. The input '${key}' is not a valid key address.", E_USER_WARNING);
            return false;
        }
    }

    /**
     * get the most recently queried table
     *
     * @access  public
     * @return  string
     */
    function getTable()
    {
        assert('is_string($this->lastSelectedTable);');
        return $this->lastSelectedTable;
    }

    /**
     * get the number of entries inside a table
     *
     * @access  public
     * @param   string        $table   (optional)
     * @param   string|array  $search  (optional)
     * @return  int
     */
    function length($table = "", $search = array())
    {
        /* 1 check input */
        if (!is_string($table)) {
            trigger_error("Wrong argument type for argument 2. String expected found '".gettype($table)."' instead.", E_USER_WARNING);
            return false;
        }
        if (!is_string($search) && !is_array($search)) {
            trigger_error("Wrong argument type for argument 2. String or array expected found '".gettype($search)."' instead.", E_USER_WARNING);
            return false;
        }

        if (!$this->_check_available()) {
            return false;
        }

        /* 2 check table parameter */
        $table = strtolower($table);
        if ($table == "" || $table == "*") {
            if (empty($this->lastSelectedTable)) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }
        if (!$this->exists($table)) {
            return 0;
        }

        /* 3 check search parameter */
        if ((is_string($search) && !empty($search)) || (is_array($search) && count($search) > 0)) {
            $search = $this->_computeSearchString($table, $search, true);
            if ($search === false) {
                return false;
            }
        } else {
            $search = "";
        }

        /* 4 query database */
        $result = $this->database->query("SELECT COUNT(*) FROM ".$this->quoteId($table). ( (!empty($search)) ? " WHERE ".$search : "" ) );

        /* 5 return result */
        if ($this->isError($result)) {
            return false;
        } else {
            $i = $result->fetchRow(DB_FETCHMODE_ORDERED,0);
            if (is_null($i) || !isset($i[0])) {
                return false;
            } else {
                return (int) $i[0];
            }
        }
    }

    /**
     * check wether a certain table has no entries
     *
     * Note: if no table is provided, the most recently used
     * table will be tested instead.
     *
     * @access  public
     * @param   string   $table  (optional)
     * @return  boolean
     */
    function isEmpty($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($table, "string");

        $table = strtolower($table);

        if ($table == "" || $table == "*") {
            if (!$this->lastSelectedTable) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        $length = (int) $this->length($table);
        return ($length == 0);
    }

    /**
     * Check wether a certain key exists
     *
     * Returns bool(true), if the adress $key (table, row, column) is defined, and otherwise bool(false).
     * If no argument is provided, the function returns bool(true) if a database connection exists and bool(false) if not.
     *
     * @uses $BufferedDbStream->exists('table.5')
     *
     * @access  public
     * @param   string  $key  adress to check
     * @return  bool
     */
    function exists($key = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($key); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($key, "string");

        $key = preg_replace("/\.(\*)?$/", "", $key);
        $key = strtolower($key);

        if ($key == "" || $this->exists == YANA_DATABASE_FAILURE) {
            if ($this->exists == YANA_DATABASE_FAILURE) {
                return false;
            } else {
                return $this->exists;
            }
        } else {
            $key = strtolower($key);
            $key = explode('.', $key);
            /* int */ $result = 0;
            if (count($key) == 1) {
                return $this->structure->isTable($key[0]);
            } else if (count($key) < 0) {
                return false;
            } else if (count($key) == 2) {
                $result = $this->database->query("SELECT 1 FROM ".$this->quoteId($key[0])." WHERE ".$this->structure->getPrimaryKey($key[0])." = ".$this->quote($key[1]));
            } else if (count($key) == 3) {
                if ($key[1] === '*') {
                    $result = $this->database->query("SELECT 1 FROM ".$this->quoteId($key[0])." WHERE ".$key[2]." IS NOT NULL");
                } else {
                    $result = $this->database->query("SELECT 1 FROM ".$this->quoteId($key[0])." WHERE ".$this->structure->getPrimaryKey($key[0])." = ".$this->quote($key[1])." AND ".$key[2]." IS NOT NULL");
                }
            } else {
                /* syntax error */
                return false;
            }

            if ($this->isError($result)) {
                return false;
            } else {
                if ($result->numRows() > 0) {
                    return true;
                } else {
                    return false;
                }
            }
        }
    }

    /**
     * check wether the current database is readonly
     *
     * @uses $BufferedDbStream->isWriteable()
     *
     * @access  public
     * @return  bool
     */
    function isWriteable()
    {
        if ($this->exists() && !$this->structure->isReadonly()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * resolve search clause
     *
     * @todo make this function accept other comparison operators like '<' and '>'
     *
     * @access  private
     * @param   string        $table
     * @param   string|array  $search
     * @param   bool          $is_table_scan   if scanning a whole table you may search for primary keys - otherwise not
     * @return  string
     * @ignore
     */
    function _computeSearchString($table, $search, $is_table_scan = false)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('preg_match("/^\w+$/s",$table);           // invalid table name, may only contain alpha-numeric characters and can not be empty');
        /* argument 2 */ assert('is_string($search) || is_array($search); // wrong argument type for argument 2, string expected');
        /* argument 3 */ assert('is_bool($is_table_scan);                 // wrong argument type for argument 3, boolean expected');

        /* optimistic type casting */
        /* argument 1 */
        settype($table,         'string');
        $table = strtolower($table);
        /* argument 2 */
        if (!is_array($search)) {
            settype($search,    'string');
        }
        /* argument 3 */
        settype($is_table_scan, 'bool');

        /* convert string to array */
        if (is_string($search)) {
            if (!preg_match("/^((, ?)?([\w\d-_\.])+=([\w\d-_\.\@ \&;])+)*$/i", $search)) {
                $this->handleError("SQL ERROR: Invalid where clause. The search string you provided contains syntax errors. Make sure it does only contain alpha-numeric characters and matches the following syntax: fieldname1=value1,fieldname2=value2");
                return false;
            }
    
            $search = preg_split("/, ?/", $search);
            settype($search, "array");
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($search as $a => $b)
            {
                $search[$a] = explode("=", $b);
                if ($this->structure->isForeignKey($table,$search[$a][0]) || $this->structure->isPrimaryKey($table,$search[$a][0])) {
                    $search[$a][2] = '=';
                } else {
                    $search[$a][1] = '%'.$search[$a][1].'%';
                    $search[$a][2] = 'like';
                }
            } /* end foreach */
            unset($a,$b);
        }

        /* resolve array */
        $result = "";
        if (count($search) > 0) {
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($search as $a => $b)
            {
                if (is_array($b) && count($b) >= 2 && count($b) <= 3 && !empty($b[0]) && isset($b[1])) {

                    /* use names that are easier to read */
                    $columnName = $b[0];
                    $value      = $b[1];
                    if (!isset($b[2])) {
                    } else {
                        $operator   = $b[2];
                    }

                    /* set input type */
                    settype($columnName, 'string');
                    settype($value,      'string');
                    settype($operator,   'string');

                    /* lower case input */
                    $columnName = strtolower($columnName);
                    $operator   = strtolower($operator);

                    /* map aliases */
                    if ($operator === '==') {
                        $operator = '=';
                    }

                    /* catch exceptions */
                    if (!in_array($operator, array('=', 'like', '<', '>', '!=', '<=', '>='))) {
                        $this->handleError("SQL ERROR: Invalid SQL statement. The operator '${operator}' is not supported.");
                        return false;
                    }
                    if (!$is_table_scan && $this->structure->isStrict() && $this->structure->isPrimaryKey($table,$columnName)) {
                        $this->handleError("SQL ERROR: Invalid SQL statement. You are trying to search for a primary key. This is not allowed, since it might cause results wether to be ambigious or empty. Try setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.");
                        return false;
                    }
                    if ($this->structure->isStrict() && !$this->structure->isColumn($table,$columnName) && !$this->structure->isForeignKey($table,$columnName)) {
                        $this->handleError("SQL ERROR: Field '".$columnName."' does not exist in table '".$table."'. Perhaps your structure file is incomplete. Try setting 'USE_STRICT' to 'false' if you want to turn this message off and continue without checking.");
                        return false;
                    }

                    /* create result */
                    $result .= ( (!empty($result)) ? "AND " : "" ).$columnName." ".$operator." ".$this->quote(addslashes($value))." ";
                }
            } /* end foreach */
            unset($a,$b);
        }

        return $result;
    }

    /**
     * import SQL from a file
     *
     * @uses $BufferedDbStream->importSQL('some_file.sql')
     *
     * @access  public
     * @return  bool
     * @name    BufferedDbStream::importsql()
     * @param   string  $sqlFile
     */
    function importSQL($sqlFile)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($sqlFile); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($sqlFile, "string");

        if ($this->_check_writeable() === true) {
            if ($this->structure->isReadonly() || (!empty($sqlFile) && !is_readable($sqlFile))) {
                return false;
            } else if (!empty($sqlFile)) {
                $raw_data = @fread(@fopen($sqlFile, 'r'), @filesize($sqlFile));
                /* Entferne Kommentare und leere Zeilen */
                $raw_data = preg_replace("/\s*\#[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/\s*\-\-[^\n]*/i", "", $raw_data);
                $raw_data = preg_replace("/;\s*\n\s*/i", "[NEXT_COMMAND]", $raw_data);
                $raw_data = preg_replace("/\s/", " ", $raw_data);

                $this->queue = array_merge($this->queue, explode("[NEXT_COMMAND]", $raw_data));
                return $this->write();
            } else {
                return true;
            }
        } else {
            return false;
        }
    }

    /**
     * get CSV string from a table
     *
     * @uses $BufferedDbStream->toString('guestbook')
     *
     * @access public
     * @param string $table (optional)
     * @return string
     */
    function toString($table = "")
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($table, "string");

        /* lower-case input */
        $table = strtolower($table);

        if ($this->exists === YANA_DATABASE_FAILURE || $this->exists === YANA_DATABASE_NOT_AVAILABLE) {
            return "No Database Connection: Database structure file not available or not loaded";
        }

        if ($table == "" || $table == "*") {
            if (empty($this->lastSelectedTable)) {
                return false;
            } else {
                $table = $this->lastSelectedTable;
            }
        }

        assert('!isset($text); /* cannot redeclare variable $text */');
        $text = "";

        if (!$this->exists($table)) {
            trigger_error("The selected table '${table}' does not exist.", E_USER_WARNING);
            return "";
        } else {

            assert('!isset($a); /* cannot redeclare variable $a */');
            $a = $this->get($table);
            if(is_array($a)) {
                assert('!isset($k); /* cannot redeclare variable $k */');
                assert('!isset($e); /* cannot redeclare variable $e */');
                foreach ($a as $k => $e)
                {
                    assert('!isset($i); /* cannot redeclare variable $i */');
                    for ($i=0; $i<count($e);$i++)
                    {
                        $b = current($e);
                        settype($b, "string");
                        $text .= $b;
                        if (next($e)) {
                              $text .= ", ";
                        }
                    } /* end for */
                    unset($i);
                    $text .= "\n";
                } /* end foreach */
                unset($k,$e);
            } /* end if */
            unset($a);
            return $text;
        }
    }

    /**
     * export database structure to a file
     *
     * Note: if the specified file already exists it will get
     * overwritten.
     *
     * Structure files need to have the extension '.config' and are
     * to be stored at the directory 'config/db/*.config'.
     *
     * @uses $BufferedDbStream->exportStructure('some_file.config')
     *
     * @access public
     * @param string $filename
     * @return boolean
     */
    function exportStructure($filename)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($filename); // wrong argument type for argument 1, string expected');

        /* optimistic type casting */
        /* argument 1 */ settype($filename, "string");

        if (!preg_match("/\.config$/", $filename)) {
            trigger_error("Unexpected file extension. The filename '${filename}' should end with '.config'.", E_USER_WARNING);
            return false;
        }
        $file = new SML($filename, CASE_UPPER);
        if (!$file->exists()) {
            $file->create();
        }
        $file->read();
        if (isset($this->structure) && $file->set($this->structure->getStructure()) && $file->write()) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * handleError
     *
     * @access  private
     * @param   mixed   $error
     * @param   mixed   $additional_info (optional)
     * @ignore
     */
    function handleError($error, $additional_info = "")
    {
        global $YANA;

        if ($this->isError($error)) {
            if (isset($YANA)) {
                $YANA->report( new log("$additional_info", $error->getMessage()) );
            } else {
                if (is_string($additional_info)) {
                    trigger_error($error->getMessage()." $additional_info", E_USER_WARNING);
                } else {
                    trigger_error($error->getMessage(), E_USER_WARNING);
                }
            }
        } else if (is_string($error)) {
            if (isset($YANA)) {
                $YANA->report( new log("${error}",$additional_info) );
            } else {
                if (is_string($additional_info)) {
                    trigger_error("$error $additional_info", E_USER_WARNING);
                } else {
                    trigger_error("$error", E_USER_WARNING);
                }
            }
        }
    }

    /**
     * buildStructure
     *
     * {@internal
     *
     * WARNING: be aware that this won't work for all DBMS since
     * the type and ammount of information provided might vary.
     * ALSO: be aware that this won't work with PEAR-DB prior
     * to version 1.6.0
     *
     * ... or in other words: avoid this function in any productive
     * environment.
     *
     * It is only there as a "last-resort" in case the structure file
     * is not there as well as for debugging, testing and reverse
     * engineering your own database to a structure file when writing
     * new plugins (and by doing so possibly making your work a bit easier).
     *
     * }}
     *
     * @access private
     * @ignore
     */
    function buildStructure()
    {
        if (!isset($this->structure)) {
            $this->structure  = new StructureFile($this->filename);
            $this->structure->init();
        }
        $tables = $this->database->getListOf('tables');
        if (!$this->isError($tables)) {
            assert('!isset($table); /* cannot redeclare variable $table */');
            foreach ($tables as $table)
            {
                $result = $this->database->query('SELECT * FROM '.$this->quoteId($table).' WHERE 1=0');
                if ($this->isError($result)) {
                    $this->handleError("DB WARNING: unable to read structure of table '${table}'.");
                } else {
                    $result = $this->database->tableInfo($result, DB_TABLEINFO_ORDER);
                    if (!is_array($result)) {
                        $this->handleError("DB WARNING: unable to read structure of table '${table}'.");
                    } else {
                        $this->structure->addStructure($table, $result);
                    }
                }
            } /* end foreach */
            unset($table);
        } else {
            $this->handleError("DB WARNING: unable to read structure of database.");
        }
    }

    /**
     * read
     *
     * @deprecated since version 2.8.3
     * @access private
     * @return boolean
     * @ignore
     */
    function read()
    {
        if ($this->exists != YANA_DATABASE_OK) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * available
     *
     * @access private
     * @return boolean
     * @ignore
     */
    function available()
    {
        if (($this->exists == YANA_DATABASE_OK && !$this->read()) || empty($this->database)) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * getFilename
     *
     * @access  protected
     * @return  string
     * @ignore
     */
    function getFilename()
    {
        return $this->filename;
    }

    /**
     * _sqlValues
     *
     * @access  protected
     * @param   array    $value
     * @return  string
     * @ignore
     */
    function _sqlValues($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_array($value); // wrong argument type for argument 1, array expected');

        /* optimistic type casting */
        /* argument 1 */ settype($value, "array");

        $k_value = array_keys($value);
        $v_value = array_values($value);

        $newfunc = create_function('&$item,$key,&$obj', '$item = $obj->quote($item);');
        array_walk($v_value, $newfunc, $this);
        $result  = '('.join(',', $k_value).')';
        $result .= ' VALUES('.join(',',$v_value).')';

        assert('is_string($result) && !empty($result);');
        return $result;
    }

    /**
     * isError
     *
     * @access  private
     * @param   mixed   $result
     * @return  boolean
     * @ignore
     */
    function isError($result)
    {
        return DB::isError($result);
    }

    /**
     * quote
     *
     * @access  private
     * @param   mixed   $value
     * @return  string
     * @ignore
     */
    function quote($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_scalar($value); // wrong argument type for argument 1, scalar expected');

        /* optimistic type casting */
        if (!is_scalar($value)) {
            trigger_error("Your SQL statement contains an unexpected non-scalar value: '".print_r($value,true)."'", E_USER_NOTICE);
            /* argument 1 */ settype($value, "string");
        }
        return $this->database->quoteSmart($value);
    }

    /**
     * smart id quoting
     *
     * implements a blacklist approach to automated quoting
     *
     * @access  private
     * @param   mixed   $value
     * @return  string
     * @ignore
     */
    function quoteId($value)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($value); // Wrong argument type for argument 1. String expected, found \''.gettype($value).'\' instead.');

        /* optimistic type casting */
        /* argument 1 */ settype($value, "string");

        if (preg_match('/\s/s', $value) || $this->_isSqlKeyword($value) === true) {
            return $this->database->quoteIdentifier($value);
        } else {
            return $value;
        }
    }

    /**
     * _check_available
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _check_available()
    {
        if ($this->available() === false) {
            trigger_error("Datebase is not available. In structure file '".$this->filename."'.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    /**
     * _check_writeable
     *
     * @access  private
     * @return  bool
     * @ignore
     */
    function _check_writeable()
    {
        if ($this->_check_available() === false) {
            return false;
        } else if (!$this->isWriteable()) {
            trigger_error("Datebase is not writeable. Check structure file '".$this->filename."' if attribute readonly is set to true.", E_USER_WARNING);
            return false;
        } else {
            return true;
        }
    }

    /**
     * build an update statement (without where clause)
     *
     * @access  private
     * @param   string  $table
     * @param   array   $values
     * @return  bool
     * @ignore
     */
    function _buildUpdate($table, $values)
    {
        /* pessimistic scanning */
        /* argument 1 */ assert('is_string($table); // wrong argument type for argument 1, string expected');
        /* argument 2 */ assert('is_array($values); // wrong argument type for argument 2, array expected');

        $sqlStmt = '';
        if (is_string($table) && !empty($table) && is_array($values)) {
            assert('!isset($row); /* cannot redeclare variable $row */');
            assert('!isset($value); /* cannot redeclare variable $value */');
            foreach ($values as $row => $value)
            {
                if (!empty($sqlStmt)) {
                    $sqlStmt .= ',';
                }
                $sqlStmt .= $row.'='.$this->quote($value);
            }
            unset($row,$value);
            $sqlStmt = 'UPDATE '.$this->quoteId(YANA_DATABASE_PREFIX.$table).' SET '.$sqlStmt. ' ';
            return $sqlStmt;
        } else {
            trigger_error('SQL syntax error.', E_USER_WARNING);
            assert('empy($sqlStmt);');
            return $sqlStmt;
        }
    }

    /**
     * returns true if $name is a known SQL keyword and false otherwise
     *
     * implements quick-search
     * + assume that the input is sorted
     * + assume that the input is upper case
     *
     * this algorithm has O(log(n)) running time
     *
     * @access  private
     * @param   string  $name
     * @return  bool
     * @ignore
     */
    function _isSqlKeyword($name)
    {
        if (!is_string($name)) {
            trigger_error("SQL syntax error. Invalid id '".gettype($name)."(".print_r($name,true).")'.", E_USER_WARNING);
            return true;
        } else if (is_numeric($name)) {
            return true;
        }

        if (is_null($this->reservedSqlKeywords)) {
            global $YANA;
            /* Load list of reserved SQL keywords (required for smart id quoting) */
            if (isset($YANA)) {
                $file = $YANA->registry->get('reserved_sql_keywords');
                $this->reservedSqlKeywords = file($file->getFilename());
            } else {
                $this->reservedSqlKeywords = array();
            }
            if (!is_array($this->reservedSqlKeywords)) {
                $this->reservedSqlKeywords = array();
            }
        } else if (empty($this->reservedSqlKeywords)) {
            return false;
        }

        $max    = count($this->reservedSqlKeywords) -1;
        $min    = 0;
        $name   = strtoupper($name);
        $n      = floor($max/2);
        $prev_n = array(-1,-1);

        while ($max > $min)
        {
            $temp = strcmp(trim($this->reservedSqlKeywords[$n]), $name);
            if ($temp > 0) {
                $max = $n;
            } else if ($temp < 0) {
                $min = $n;
            } else {
                return true;
            }

            array_shift($prev_n);
            $prev_n[count($prev_n)] = $n;

            if ($min!=$n) {
                $n = $min + floor( ($max-$min) /2 );
            } else {
                $n = $min + ceil( ($max-$min) /2 );
            }

            if ($prev_n[0]==$n||$prev_n[1]==$n) {
                return false;
            }

        } /* end while */
        return false;
    }

    /**
     * return when the row was last modified
     *
     * @access  private
     * @param   string  $key
     * @return  int
     * @ignore
     */
    function _lastModified($key)
    {
        assert('is_string($key);');
        settype($key, 'string');
        global $YANA;

        $file_exists = false;            

        /* load file */
        if (is_null($this->lastModified)) {
            if (isset($YANA)) {
                $this->lastModified = new SML($YANA->getVar('TEMPDIR').'db_last_modified.tmp', CASE_UPPER);
            } else {
                $this->lastModified = new SML('cache/'.'db_last_modified.tmp', CASE_UPPER);
            }
            $this->lastModified->init();
            if (!$this->lastModified->exists()) {
                $this->lastModified->create();
                $file_exists = false;
            } else {
                $file_exists = true;
            }
        }

        $keyArray = explode('.', $key);

        /* get time */
        if ($file_exists === true) {
            if (count($keyArray) >= 2) {
                $lastModified = $this->lastModified->get($keyArray[0].'.'.$keyArray[1]);
            } else if (count($keyArray) === 1) {
                $array = $this->lastModified->get($keyArray[0]);
                sort($array);
                $lastModified = array_pop($array);
            } else {
                return 0;
            }
        } else {
            $lastModified = 0;
        }

        /* update time */
        if (($file_exists === false || $lastModified != time()) && count($keyArray) >= 2) {
            $this->lastModified->insert(array($keyArray[0] => array($keyArray[1]=>time())));
        }

        /* return data */
        if ($lastModified > time()) {
            return 0;
        } else if (!empty($lastModified)) {
            return (int) $lastModified;
        } else {
            return 0;
        }
    }

    /**
     * Reset the object to default values
     *
     * Resets the history for the last selected table,
     * resets the queue of pending SQL statements and
     * resets the database cache.
     *
     * @access  public
     */
    function reset()
    {
        $this->lastSelectedTable   = "";
        $this->queue               = array();
        $this->cache               = array();
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $another_object
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file and use equal database connections.
     *
     * @access   public
     * @param    object $another_object
     * @return   string
     */
    function equals($another_object)
    {
        if (is_object($another_object) && method_exists($another_object,'getClass') && $another_object->getClass() === $this->getClass()) {
            if ($this->filename === $another_object->filename && $this->database == $another_object->database) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}

?>