<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * counter wrapper
 *
 * This class implements persistent counters - optional with IP checking -
 * for various purposes: e.g. visitor counts, statistics, polls.
 *
 * @access      public
 * @package     yana
 * @subpackage  utilities
 */
class Counter extends Object
{
    /**#@+
     * @access private
     * @ignore
     */

    /** @var string  */ var $directory = "";
    /** @var SML     */ var $file      = null;
    /** @var SML     */ var $ip_file   = null;
    /** @var boolean */ var $use_ip    = true;

    /**#@-*/

    /**
     * create a new instance
     *
     * Use the $use_ip parameter with the setting YANA_COUNTER_USE_IP
     * to create a counter with IP checking (does'nt change on reload)
     * or with the setting YANA_COUNTER_IGNORE_IP to create a counter
     * without IP checking (always changes when an update is called).
     *
     * @param string $filename (optional)
     * @param int $use_ip (optional) default = YANA_COUNTER_USE_IP
     */
    function Counter($filename = "", $use_ip = YANA_COUNTER_USE_IP)
    {
        settype($filename, "string");
        global $YANA;

        $filename = strtolower(preg_replace("/[^\w\d_]/", "", $filename));
        if (empty($filename)) {
            /* default value */
            $filename = "statistik";
        }

        if (isset($YANA)) {
            $this->directory = $YANA->getVar("COUNTERDIR");
        } else {
            $this->directory = 'config/counter/';
        }
        $this->file    = new SML($this->directory.$filename.".cnt", CASE_UPPER);
        $this->ip_file = new SML($this->directory.$filename.".ip", CASE_UPPER);
        $this->use_ip  = ($use_ip === YANA_COUNTER_USE_IP);

        if ($this->use_ip && $this->ip_file->exists() && filemtime($this->ip_file->getFilename()) + 10000 < time()) {
            /* flush old data */
            $this->ip_file->read();       /* read */
            $this->ip_file->set(array()); /* flush */
            $this->ip_file->write();      /* write empty file */
        }
    }

    /**
     * get counter
     *
     * Returns the current counter settings as:
     *
     * <pre>
     * Array(
     *    'COUNT' => int
     *    'INFO'  => string
     * )
     * </pre>
     *
     * Note: you may use the wildcard '*' to get
     * all counters as an associative, multidimensional array:
     *
     * <pre>
     * Array(
     *    ID1 => Array(
     *        'COUNT' => int
     *        'INFO'  => string
     *    )
     *    ID2 => Array(
     *        'COUNT' => int
     *        'INFO'  => string
     *    )
     *    ...
     * )
     * </pre>
     *
     * @access public
     * @param string $id (optional)
     * @return array
     */
    function get($id = "*")
    {
        settype($id, "string");
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) {
            /* syntax error */
            return false;
        }

        if (empty($id) || $id == "*") {
            /* return all */
            return $this->file->get();
        } else {
            /* return single entry */
            return $this->file->get($id);
        }
    }

    /**
     * get counter info
     *
     * The 'info' field is an optional text value, that
     * describes the counter.
     *
     * This function returns the current description of counter $key.
     *
     * Note: you may use the wildcard '*' to get
     * an associative array for all counters:
     *
     * <pre>
     * Array(
     *    ID1 => string
     *    ID2 => string
     *    ...
     * )
     * </pre>
     *
     * @access public
     * @param string $id (optional)
     * @return mixed
     */
    function getInfo($id = "*")
    {
        settype($id, "string");
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) {
            /* syntax error */
            return false;
        }

        if (empty($id) || $id == "*") {
            /* return all */
            $A = $this->file->get();
            $B = array();
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ( $A as $a => $b )
            {
                if(!empty($b['INFO'])) {
                    $B[$a] = $b['INFO'];
                }
            }
            unset($a,$b);
            asort($B, SORT_STRING);
            return $B;
        } else {
            /* return single entry */
            return $this->file->get("$id.INFO");
        }
    }

    /**
     * get counter value
     *
     * This function returns the current value of counter $key.
     *
     * Note: you may use the wildcard '*' to get
     * an associative array for all values:
     *
     * <pre>
     * Array(
     *    ID1 => int
     *    ID2 => int
     *    ...
     * )
     * </pre>
     *
     * @access public
     * @param string $id (optional)
     * @return mixed
     */
    function getCount($id = "*")
    {
        settype($id, "string");
        if (!preg_match("/^([\w\d_]+|\*)$/i", $id)) return false;

        if (empty($id) || $id == "*") {
            /* return all */
            $A = $this->file->get();
            $B = array();
            assert('!isset($a); /* cannot redeclare variable $a */');
            assert('!isset($b); /* cannot redeclare variable $b */');
            foreach ($A as $a => $b)
            {
                $B[$a] = $b['COUNT'];
            }
            unset($a,$b);
            asort($B, SORT_NUMERIC);
            return $B;
        } else {
            /* return single entry */
            return $this->file->get("$id.COUNT");
        }
    }

    /**
     * Increment/Decrement counter
     *
     * Adds $ammount (defaults to +1) to the counter $id and sets
     * the counter description (the 'info' field)
     * to $info (defaults to "").
     *
     * Note: as you might already have guessed, using a negative
     * value for $ammount decrements the counter.
     *
     * @access public
     * @param string $id
     * @param string $info (optional)
     * @param int $ammount (optional) default = 1
     * @return int
     */
    function count($id, $info = "", $ammount = 1)
    {
        settype($id, "string");
        settype($info, "string");
        settype($ammount, "integer");

        if (!preg_match("/^[\w\d_]+$/i", $id)) {
            /* syntax error */
            return false;
        }

        /* BEGIN count */
            if (!$this->file->exists()) {
                /* create on first access */
                $this->file->create();
            }
            $this->file->read();
            $NEW_COUNTER                = array();
            $NEW_COUNTER[$id]           = $this->file->get($id);
            settype($NEW_COUNTER[$id]['COUNT'], "integer");
            $NEW_COUNTER[$id]['COUNT'] += $ammount;
            if (!empty($info)) {
                /* add info when avialable */
                $NEW_COUNTER[$id]['INFO'] = $info;
            }
        /* END count */

        if ($this->use_ip) {
            /* ip */
            if (!$this->ip_file->exists()) {
                /* create on first access */
                $this->ip_file->create();
            }
            $this->ip_file->read();
            if (isset($_SERVER['REMOTE_ADDR'])) {
                $ip           = preg_replace("/[^\w\d_]/", "_", $_SERVER['REMOTE_ADDR']);
            } else {
                $ip           = '0.0.0.0';
            }
            $last_visit       = (int) $this->ip_file->get("$id.$ip");
            $NEW_IP           = array();
            $NEW_IP[$id][$ip] = time() + 10000;
        }

        if (!$this->use_ip || $last_visit < time()) {
            /* if ip is ignored or no recent entry for this ip found */
            $this->file->insert($NEW_COUNTER);

            if ($this->file->write()) {
                /* write changes to file */
                if ($this->use_ip) {
                    /* dont write ip to file if ip is ignored */
                    $this->ip_file->insert($NEW_IP);
                    $this->ip_file->write();
                }
                return $NEW_COUNTER[$id]['COUNT'];
            }
        }

        return $this->getCount($id);

    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $another_object
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class, they
     * both refer to the same filesystem resource and use
     * the same IP settings.
     *
     * @access   public
     * @param    object $another_object
     * @return   string
     */
    function equals($another_object)
    {
        if (is_object($another_object) && method_exists($another_object,'getClass') && $another_object->getClass() === $this->getClass()) {
            if ($this->file->equals($another_object->file) && $this->use_ip === $another_object->use_ip) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}

?>