<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * Manipulate a directory
 *
 * This class represents a directory.
 * You may use this to get a list of contents, or remove,
 * or create a directory.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class DirStream extends InputStream
{

    /**#@+
     * @ignore
     * @access  protected
     */
    /** @var array  */ var $content = array();
    /** @var array  */ var $options = array();
    /** @var string */ var $filter  = "";
    /**#@-*/

    /**
     * read contents and put results in cache (filter settings will be applied)
     *
     * @uses $DirStream->read()
     * @access public
     * @return boolean
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = dirlist($this->getFilename(), $this->filter);
            sort($this->content);
            $this->options = $this->attributes;
            return true;
        }
    }


    /**
     * return list of files within the directory
     *
     * NOTE: will only return filenames with the path stripped
     *
     * @uses    $DirStream->get()
     * @access  public
     * @return  array
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return $this->content;
        } else {
            return array();
        }
    }

    /**
     * create this directory
     *
     * Tries to create the directory.
     * Check the developer's cookbook for an example
     * to this function.
     *
     * @uses $DirStream->create(660)
     *
     * @access  public
     * @return  bool
     * @name    DirStream::create()
     * @param   int $mode access mode
     */
    function create($mode = 777)
    {
        assert('is_int($mode);');
        if (!is_int($mode) || $this->exists() || !isset($this->filename)) {
            return false;
        } else if (!preg_match('/^[1-7]{1}[0-7]{2}$/',"$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        } else {
            return @mkdir($this->filename,"0$mode");
        }
    }

    /**
     * remove this directory
     *
     * By option you may choose to also recursivly remove all files and subdirectories inside.
     * Otherwise the directory will only be removed if it is empty.
     *
     * @uses    $DirStream->delete(false)
     *
     * @access  public
     * @param   boolean  $is_recursive  triggers wether to remove directories even if they are not empty, default = false
     * @return  boolean
     */
    function delete($is_recursive=false)
    {
        if ($is_recursive === true) {
            $content = dirlist($this->filename, '');
            assert('!isset($key); /* cannot redeclare variable $key */');
            assert('!isset($element); /* cannot redeclare variable $element */');
            foreach ($content as $a => $b)
            {
                $b = '/'.$b;
                if (is_file($this->filename.$b)) {
                    $file = new secureFileStream($this->filename.$b);
                    if ($file->delete() === false) {
                        return false;
                    }
                } elseif (is_dir($this->filename.$b)) {
                    $dir = new DirStream($this->filename.$b);
                    if ($dir->delete(true) === false) {
                        return false;
                    }
                } else {
                    /* intentionally left blank */
                }
            } /* end foreach */
            unset($a,$b);
        }
        if (@rmdir($this->getFilename()) === false) {
            trigger_error("Unable to delete directory '".$this->getFilename()."'.", E_USER_NOTICE);
            exit(1);
            return false;
        } else {
            return true;
        }
    }

    /**
     * return a string representation of this directory
     *
     * Returns a string with the contents of this directory
     *
     * @uses    $DirStream->toString()
     *
     * @access  public
     * @return  string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Directory ".$this->filename." does not exist\n";
        } else if ($this->isEmpty()) {
            return "Directory ".$this->filename." is not loaded\n";
        } else {
            return implode("\n", $this->get());
        }
    }

    /**
     * check wether the directory has no contents
     *
     * Returns bool(true) if there are no files that
     * match the current filter and bool(true) if there
     * is at least 1 file that matches.
     *
     * @uses    $DirStream->isEmpty()
     *
     * @access  public
     * @return  boolean
     */
    function isEmpty()
    {
        return empty($this->content);
    }

    /**
     * get the number of files inside the directory
     *
     * This returns a positive integer.
     * Note that this functions counts the files in respect
     * to the currently set file filter. So the number
     * of files reported here and the number in total
     * may vary.
     *
     * @uses $DirStream->length()
     *
     * @access public
     * @return int
     */
    function length()
    {
        return count($this->content);
    }

    /**
     * list contents of a directory
     *
     * The argument $filter may contain multiple file extension,
     * use a pipe '|' sign to seperate them.
     * Example: "*.xml|*.html" will find all xml- and html-files
     *
     * @uses $DirStream->dirlist('*.cfg')
     *
     * @access  public
     * @param   string  $filter
     * @return  array
     */
    function dirlist($filter)
    {
        assert('is_string($filter);');
        settype($filter, "string");
        $this->filter = $filter;
        $this->read();
        return $this->content;
    }

    /**
    * getContent
    *
    * @uses $DirStream->getContent()
    *
    * @access public
    * @return array
    * @ignore
    */
    function getContent()
    {
         if ($this->isEmpty()) {
             $this->read();
         }
         $REF = array();
         for ($i=0; $i<count($this->content); $i++)
         {
             $REF[$i]['VALUE'] = $this->content[$i];
             $REF[$i]['NAME']  = substr($this->content[$i], 0, strrpos($this->content[$i], "."));
         }
         return $REF;
    }

    /**
     * reset statistic data
     *
     * @uses $DirStream->resetStats()
     *
     * @access protected
     * @ignore
     */
    function resetStats()
    {
        if (!preg_match('/.*\/$/', $this->filename)) {
            $this->filename .= '/';
        }
        if ($this->exists()) {
            clearstatcache();
            $this->attributes['MTIME'] = filemtime($this->filename);
            clearstatcache();
        }
    }

    /**
     * check if directory exists and is readable
     *
     * @access public
     * @return boolean
     */
    function exists()
    {
        return is_dir($this->filename) && file_exists($this->filename) && is_readable($this->filename);
    }

}

?>