<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * @ignore
 */

if (!defined('DB_OK')) {
    /**
     * @ignore
     */
    define('DB_OK', 1);
}
if (!defined('DB_ERROR_UNSUPPORTED')) {
    /**
     * @ignore
     */
    define('DB_ERROR_UNSUPPORTED', 0);
}

/**
 * FileDb_connection
 *
 * Mapper for sql statements to datfile commands.
 * It implements only a required subset of the interface
 * of PEAR-DB as needed by the bufferedDbStream class
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 *
 * @ignore
 */
class FileDb_connection extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var Counter       */ var $auto_increment = null;
    /** @var StructureFile */ var $structure      = null;
    /** @var string        */ var $database       = "";
    /** @var string        */ var $table          = "";
    /** @var array         */ var $src            = array();
    /** @var array         */ var $idx            = array();
    /** @var string        */ var $sort           = "";
    /** @var bool          */ var $desc           = false;
    /** @var bool          */ var $auto_commit    = true;
    /** @var string        */ var $BASE_DIR       = 'config/db/';
    /** @var array         */ var $cache          = array();

    /**#@-*/

    /**
     * constructor
     *
     * @param string $database (optional)
     */
    function FileDb_connection($database)
    {
        settype($database, "string");
        global $YANA;

        $database = strtolower(preg_replace('/([\w\d_]+)\.config$/', '${1}', $database));
        $database = preg_replace('/[\.\*]/', '', $database);

        if (isset($YANA)) {
            $this->BASE_DIR = $YANA->getVar('DBDIR');
        }

        if (!empty($database)) {
            if (!is_dir($this->BASE_DIR . $database)) {
                mkdir($this->BASE_DIR . $database, 0600);
            }
            if (is_readable($this->BASE_DIR . $database)) {
                $this->database = $database;
                $this->src[$database] = array();
                $this->idx[$database] = array();
                $this->cache = array();
            }
        }

    }

    /**
     * autoCommit
     *
     * @access public
     * @param boolean $mode (optional) default = false
     * @return int
     */
    function autoCommit($mode=false)
    {
        settype($mode, 'boolean');
        $this->auto_commit = $mode;
        return DB_OK;
    }

    /**
     * rollback
     *
     * @access public
     * @return int
     */
    function rollback()
    {
        $this->cache = array();
        $this->src[$this->database][$this->table]->reset();
        $this->idx[$this->database][$this->table]->rollback();
        return DB_OK;
    }

    /**
     * commit
     *
     * @access public
     * @return boolean
     */
    function commit()
    {
        return $this->_write(true);
    }

    /**
     * getListOf
     *
     * @access public
     * @param string $type valid values for $type  are tables, databases
     * @return array
     */
    function getListOf($type)
    {
        settype($type, 'string');
        switch($type)
        {
            case 'tables':
                return $this->structure->getTables();
            break;
            case 'databases':
                $databases = dirlist($this->BASE_DIR,'*.config');
                for ($i=0;$i<count($databases);$i++)
                {
                    $databases[$i] = preg_replace('/\.config$/i', '', $databases[$i]);
                }
                return $databases;
            break;
            default:
                return DB_ERROR_UNSUPPORTED;
            break;
        }
    }

    /**
     * limitQuery
     *
     * @access public
     * @param string $sqlStmt
     * @param int $offset (optional) default = 0
     * @param int $limit (optional) default = 0
     * @return filedb_result
     */
    function limitQuery($sqlStmt, $offset=0, $limit=0)
    {
        settype($sqlStmt, "string");
        settype($offset,  "integer");
        settype($limit,   "integer");

        $sqlStmt = str_replace("\n", '\n', $sqlStmt);

        /* this is meant for debugging */
/*         print '<pre>'.$sqlStmt.'</pre>'; */

        /* caching */   
        if (isset($this->cache[$sqlStmt])) {
            return new filedb_result($this->cache[$sqlStmt]);
        }

        if (preg_match('/^select\s1\sfrom\s\"([\w\d_]+)\"\swhere\s(.*?)\s*;?$/si', $sqlStmt, $m)) {
            $from       = $this->_select( @ $m[1]);
            $where      = @ $m[2];

            if ($from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                if ($this->_isset($from, $where)) {
                    $result = array(1);
                    $this->cache[$sqlStmt] = $result;
                    return new filedb_result($result);
                } else {
                    return new filedb_result(array());
                }
            }

        } else if (preg_match('/^select\scount\(\*\)\sfrom\s\"([\w\d_]+)\"(?:\swhere\s(.*?))?\s*;?$/si', $sqlStmt, $m)) {

            /* parameter 1 */
            $from = $this->_select($m[1]);
            if ($from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

            /* parameter 2 */
            if (!empty($m[2])) {
                $where = $this->_where($m[2]);
                if ($where === false) {
                    return new filedb_result(null, 'SQL ERROR: Syntax error.');
                }
            } else {
                $where = array();
            }

            /* look up */
            $length = $this->_length($from,$where);

            /* create result */
            if (is_int($length)) {
                $result = array(array(0=>$length));
                $this->cache[$sqlStmt] = $result;
                return new filedb_result($result);
            } else {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

        /*
         * NOTE:
         * SUPPORT FOR TABLE JOINS IS STILL EXPERIMENTAL AND VERY INCOMPLETE.
         * THIS FEATURE IS TO ARRIVE WITH VERSION 2.8.6 - USE IT AT YOUR OWN RISK!
         *
         * For now let's limit this to requests for 2 tables, no additional where clause
         * and "select * ...".
         * That's not much - but, still it's a start!
         */

        } else if (preg_match('/^select\s(\*)\sfrom\s((?:\"[\w\d_]+\"(?:,\s)?){2})\swhere\s([\w\d_]+\.[\w\d_]+\s?=\s?[\w\d_]+\.[\w\d_]+)(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/si', $sqlStmt, $m)) {

            /*
             * Important note: this API does NOT allow cross-joins.
             * The 'where'-clause is mandatory here.
             */

            $what       = trim(strtolower(@ $m[1]));
            $from       = $m[2];
            $from       = str_replace('"', '', $from);
            $from       = preg_split('/,\s?/',$from);

            $where      = $m[3];
            $where      = preg_split('/\s?=\s?/',$where);

            $where[0]   = preg_split('/\./',$where[0]);
            $where[1]   = preg_split('/\./',$where[1]);

            /* table scan on table A */
            $this->table = strtolower(trim($where[0][0]));
            $resultSet   = $this->_get($what);

            /* fetch matching entries from table B */
            $this->table = strtolower(trim($where[1][0]));
            for ($i=0; $i<count($resultSet); $i++)
            {
                $resultSetB = $this->_get($what, array($where[1][1],$resultSet[$i][strtoupper($where[0][1])],'='));
                if (is_array($resultSetB)) {
                    $resultSet[$i] = mergeArrays($resultSet[$i], $resultSetB);
                } else {
                    unset($resultSet[$i]);
                }
            }

            if (!empty($resultSet)) {
                $this->cache[$sqlStmt] = $resultSet;
                return new filedb_result($resultSet);
            } else {
                $this->cache[$sqlStmt] = array();
                return new filedb_result(array());
            }

        } else if (preg_match('/^select\s(.*?)\sfrom\s\"([\w\d_]+)\"(?:\swhere\s(.*?))?(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/si', $sqlStmt, $m)) {

            /* SELECT */
            $what       = trim(strtolower(@ $m[1]));
            $from       = $this->_select( @ $m[2]);
            if (!empty($m[3])) {
                $where  = $this->_where($m[3]);
            } else {
                $where  = array();
            }
            $this->sort = $this->_order_by(@$m[4]);
            $this->desc = ! @empty(         $m[5]);

            if ($where === false || $from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                $result = $this->_get($what, $where, $offset, $limit);
                $this->cache[$sqlStmt] = $result;
                return new filedb_result($result);
            }

        } else if (preg_match('/^insert into\s\"([\w\d_]+)\"\s*\((.*?)\)\s+VALUES\((.*?)\)\s*;?$/si', $sqlStmt, $m)) {

            /* INSERT */
            $from  = $this->_select(@ $m[1]);
            if ($from === false) {
                return new filedb_result(null, "SQL ERROR: Table '".(@$m[1])."' is unknown.");
            }
            if (!preg_match("/^\".*?(\",\".*?)*\"$/", $m[3])) {
                trigger_error("SQL syntax error. The string '".print_r($m[3], true)."' is not a valid value in the provided SQL-statement.", E_USER_WARNING);
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                assert('!isset($keys);');
                assert('!isset($values);');
                $keys = explode(',',@$m[2]);
                $values = explode('","', substr(@ $m[3], 1, -1));
                $set   = $this->_set($keys, $values);
                if (empty($set)) {
                    return new filedb_result(null, 'SQL ERROR: The statement contains illegal values.');
                } else {
                    $set   = $this->_increment($set);
                }
                unset($keys,$values);
            }

            $smlfile =& $this->src[$this->database][$this->table];
            $idxfile =& $this->idx[$this->database][$this->table];

            $primary_id  = strtoupper($this->structure->getPrimaryKey($this->table));
            $primary_key = $set[$primary_id];
            unset($set[$primary_id]);

            /* create column */
            assert('!isset($column); /* cannot redeclare variable $column */');
            foreach ($this->structure->getColumns($this->table) as $column)
            {
                $column = strtoupper($column);
                if (isset($set[$column]) && $this->structure->getType($this->table,$column) == 'array') {
                    $set[$column] = SML::getFile(explode('\n',$set[$column]), CASE_UPPER);
                }
                if (isset($set[$column])) {
                    /* check unique constraint */
                    if ($this->structure->isUnique($this->table,$column) === true) {
                        assert('!isset($tmp); /* cannot redeclare variable $tmp */');
                        $tmp = $idxfile->get($column,$set[$column]);
                        if (!empty($tmp)) {
                            return new filedb_result(null, "SQL ERROR: Cannot insert entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                        }
                        unset($tmp);
                    }
                    /* update indexes */
                    if ($this->structure->hasIndex($this->table,$column) === true || $this->structure->isUnique($this->table,$column) === true) {
                        $idxfile->create($column,array($primary_key,$set[$column]));
                    }
                }
            }
            unset($column);
            $set = array($primary_id => array($primary_key => $set));
            if (is_array($smlfile->get($primary_id.'.'.$primary_key))) {
                return new filedb_result(null, "SQL ERROR: Cannot insert entry with primary key = ${primary_key}. Another entry with this key already exists.");
            } else {
                if ($smlfile->insert($set)) {
                    return $this->_write();
                } else {
                    return new filedb_result(null, "unable to save changes");
                }
            }

        } else if (preg_match('/^update\s\"([\w\d_]+)\"\sset\s((?:\s*\w+\s*\=\s*".*?"\s*,?)+)(?:\s+where\s(.*))?\s*;?$/si', $sqlStmt, $m)) {

            /* UPDATE */
            $from  = $this->_select(   @$m[1]);
            if ($from === false) {
                return new filedb_result(null, "SQL ERROR: Table '".(@$m[1])."' is unknown.");
            }
            assert('!isset($nr_of_hits) && !isset($m1);');
            $nr_of_hits = preg_match_all('/\s*(\w+)\s*\=\s*"(.*?)"\s*,?/s', $m[2], $m1);
            if ($nr_of_hits === false || $nr_of_hits < 1) {
                trigger_error("SQL syntax error. The string '".print_r($m[2], true)."' is not a valid value in the provided SQL-statement.", E_USER_WARNING);
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            } else {
                unset($nr_of_hits);
                $set   = $this->_set($m1[1], $m1[2]);
                unset($m1);
            }
            if (!empty($m[3])) {
                $where  = $this->_where($m[3]);
            } else {
                $where  = array();
            }
            if ($where === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

            assert('!isset($primary_key);');
            $primary_key = strtoupper($this->structure->getPrimaryKey($this->table));

            assert('!isset($column); /* cannot redeclare variable $column */');
            foreach ($this->structure->getColumns($this->table) as $column)
            {
                $column = strtoupper($column);
                if ($this->structure->getType($this->table,$column) == 'array') {
                    if (!empty($set[$column])) {
                        $set[$column] = SML::decode($set[$column], CASE_UPPER);
                    } else {
                        $set[$column] = array();
                    }
                }
                /* check unique constraint */
                if ($this->structure->isUnique($this->table,$column) === true) {
                    assert('!isset($tmp); /* cannot redeclare variable $tmp */');
                    $tmp = $idxfile->get($column,$set[$column]);
                    if (!empty($tmp)) {
                        return new filedb_result(null, "SQL ERROR: Cannot update entry with column ${column} = ".$set[$column].". The column has an unique constraint and another entry with the same value already exists.");
                    }
                    unset($tmp);
                }
            } /* end foreach */
            unset($column);

            /* extract new primary key */
            assert('!isset($rule); /* cannot redeclare variable $rule */');
            foreach ($where as $rule)
            {
                if (strcasecmp($rule[0], $primary_key) == 0) {
                    $key = strtoupper($rule[1]);
                    break;
                }
            }
            unset($rule);

            $smlfile =& $this->src[$this->database][$this->table];
            $idxfile =& $this->idx[$this->database][$this->table];

            /* if primary key is renamed, the old one has to be replaced */
            if (isset($set[$primary_key]) && $key != $set[$primary_key]) {
                $smlfile->remove("$primary_key.$key");
                $temp = $set[$primary_key];
                $set  = array($temp => $set);
                unset($set[$temp][$primary_key]);
                unset($temp);
            } else if (!empty($key)) {
                $set = array($key => $set);
            } else {
                return new filedb_result(null, "SQL ERROR: Syntax error $sqlStmt");
            }
            $set = array($primary_key => $set);

            if ($smlfile->insert($set)) {
                /* after data has been changed, reorganize all indexes */
                $idxfile->create();
                return $this->_write();
            } else {
                return new filedb_result(null, "unable to save changes");
            }

        } else if (preg_match('/^delete\s+from\s\"([\w\d_]+)\"\swhere\s(.*)(?:\s*order by\s([\w\d_",]+))?(\s*desc)?\s*;?$/Usi', $sqlStmt, $m)) {
            /* DELETE */

            $from       = $this->_select( @ $m[1]);
            if (!empty($m[2])) {
                $where  = $this->_where($m[2]);
            } else {
                $where  = array();
            }
            $this->sort = $this->_order_by(@$m[3]);
            $this->desc = ! @empty(         $m[4]);

            if ($where === false || $from === false) {
                return new filedb_result(null, 'SQL ERROR: Syntax error.');
            }

            $smlfile =& $this->src[$this->database][$this->table];
            $idxfile =& $this->idx[$this->database][$this->table];

            $i = $this->_get($this->structure->getPrimaryKey($this->table), $where, 0, 1);

            if (empty($i)) {
                /* error */
                return new filedb_result(null, "SQL-ERROR: unable to delete; entry does not exist");
            } else {
                $i = $i[0][strtoupper($this->structure->getPrimaryKey($this->table))];
                if ($smlfile->remove($this->structure->getPrimaryKey($this->table).'.'.$i)) {
                    /* after data has been changed, reorganize all indexes */
                    $idxfile->create();
                    return $this->_write();
                } else {
                    return new filedb_result(null, "unable to save changes");
                }
            }

        } else {
            /* throw a line if unknown statement encountered */

            global $YANA;

            if (isset($YANA)) {
                $YANA->setVar("__PHP_DBASE_ERROR__", "Ress#id: @74a34b2b616b34f8eb3f5b2fbce7187c");
            }

            trigger_error("Invalid or unknown SQL statement: $sqlStmt.",E_USER_ERROR);
            return new filedb_result(null, "SQL ERROR: Syntax error '$sqlStmt'.");
        }
    }

    /**
     * query
     *
     * alias of limitQuery() with $offset and $limit params stripped
     *
     * @access public
     * @param string $sqlStmt
     * @return filedb_result
     */
    function query($sqlStmt)
    {
        assert('is_string($sqlStmt);');
        settype($sqlStmt, "string");
        return $this->limitQuery($sqlStmt);
    }

    /**
     * setStructure
     *
     * @access public
     * @param StructureFile &$structure
     */
    function setStructure(&$structure)
    {
        /* this is a simple work-around to use PHP5-like type-checking in PHP4 */
        assert('is_a($structure, "StructureFile");');
        $this->structure =& $structure;
    }

    /**
     * _sort
     *
     * @access private
     * @param array $a
     * @param array $b
     * @return int
     * @ignore
     */
    function _sort($a, $b, $columns = null)
    {
        if (is_null($columns)) {
            $columns = $this->sort;
        }
        if (!is_array($columns) || count($columns) === 0) {
            return 0;
        } else {
            $sort = array_shift($columns);

            if (!is_array($a) || !is_array($b)) {
                return 0;
            } else if (!isset($a[$sort]) && !isset($b[$sort])) {
                $result = $this->_sort($a,$b,$columns);
            } else if (!isset($a[$sort])) {
                $result = -1;
            } else if (!isset($b[$sort])) {
                $result = 1;
            } else if ($a[$sort]<$b[$sort]) {
                $result = -1;
            } else if ($a[$sort]>$b[$sort]) {
                $result = 1;
            } else {
                if (count($columns) > 0) {
                    $result = $this->_sort($a,$b,$columns);
                } else {
                    $result = 0;
                }
            } /* end if */
            assert('is_int($result) && $result >= -1 && $result <= 1; // unexpected result');

            if ($this->desc) {
                return - $result;
            } else {
                return $result;
            }
        } /* end if */
    }

    /**
     * _select
     *
     * loads a table and returns table name
     *
     * NOTE: returns boolean(false) on error
     *
     * @access private
     * @param  string $select
     * @return string
     * @ignore
     */
    function _select($select)
    {
        assert('is_string($select);');
        settype($select, 'string');
        $select = strtolower(trim($select));

        if (empty($select)) {
            return false;
        } else if (!isset($this->src[$this->database][$select])) {
            $this->table = $select;
            assert('!isset($database); // Cannot redeclare $database');
            $database = $this->structure->getAssociation($this->table);
            if (is_null($database)) {
                $database = $this->database;
            }
            assert('is_string($database);');
            if ($this->structure->isStrict() && !$this->structure->isTable($this->table)) {
                trigger_error("There is no table '".$this->table."' in current structure file.", E_USER_WARNING);
                return null;
            }
            $filename = $this->BASE_DIR.$database.'/'.$this->table;
            $smlfile = new SML($filename.'.sml', CASE_UPPER);
            unset($database);
            $smlfile->init();
            if (!$smlfile->exists()) {
                $smlfile->create();
                $filename = $this->BASE_DIR.$this->database.'/'.$this->table;
                $smlfile = new SML($filename.'.sml', CASE_UPPER);
            }
            assert('is_string($this->database) && !empty($this->database);');
            assert('is_string($this->table) && !empty($this->table);');
            $this->src[$this->database][$this->table] =& $smlfile;
            $this->idx[$this->database][$this->table] =& new FileDb_index($this->structure,$this->table,$smlfile,$filename.'.idx');

            return $this->table;
        } else {
            $this->table = $select;
            assert('is_string($this->table);');
            return $this->table;
        }
    }

    /**
     * _where
     *
     * NOTE: returns boolean(false) on error
     *
     * @access private
     * @param  string $where
     * @param  string $table
     * @return array|bool(false)
     * @ignore
     */
    function _where($where)
    {
        assert('is_string($where);');
        settype($where, "string");

        $where = explode(' AND ', trim($where));
        if (empty($where)) {
            return array();
        } else {
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($val); /* cannot redeclare variable $val */');
            foreach($where as $i => $val)
            {
                if (!empty($val)) {
                    if (preg_match('/^\s*(\S+)\s?(=|LIKE|like|<|>|\!=|<=|>=)\s?\"(.+)\"\s*;?$/', $val, $m)) {
                        $where[$i]    = array();
                        $where[$i][0] = strtoupper($m[1]);
                        $where[$i][1] = toRegExp($m[3]);
                        $m[2] = strtolower($m[2]);
                        if ($m[2] === 'like') {
                            $where[$i][1] = str_replace('%','.*',$where[$i][1]);
                            $where[$i][1] = str_replace('_','.?',$where[$i][1]);
                        }
                        $where[$i][2] = $m[2];
                    } else {
                        /* syntax error */
                        return false;
                    }
                }
            } /* end foreach */
            unset($i, $val);
            return $where;
        }
    }

    /**
     * _order_by
     *
     * NOTE: returns boolean(false) on error
     *
     * @access private
     * @param  string $order_by
     * @return array
     * @ignore
     */
    function _order_by($order_by)
    {
        assert('is_string($order_by) || is_null($order_by);');
        if (empty($order_by)) {
            return array();
        } else {
            settype($order_by, "string");
    
            $order_by = explode(',', strtoupper(trim($order_by)));
            $result = array();
            if (empty($order_by)) {
                return array();
            } else {
                assert('!isset($m); /* cannot redeclare variable $m */');
                assert('!isset($val); /* cannot redeclare variable $val */');
                foreach($order_by as $val)
                {
                    /* unquote id */
                    if (!empty($val) && preg_match('/^\s*"(.*)"\s*$/',$val,$m)) {
                        array_push($result,$m[1]);
                    }
                } /* end foreach */
                unset($val);
                return $result;
            } /* end if */
        } /* end if */
    }

    /**
     * _set
     *
     * @access private
     * @param array $keys
     * @param array $values
     * @return array
     * @ignore
     */
    function _set($keys, $values)
    {
        assert('is_array($keys);');
        assert('is_array($values);');
        if (!is_array($keys) || !is_array($values)) {
            return array();
        }

        assert('count($keys) == count($values);');
        /* array_combine() became available in PHP 5, here's a code to simulate this behaviour for PHP 4 */
        $insert = array();
        assert('!isset($i); /* Unable to redeclare $i */');
        assert('!isset($isStrict); /* Unable to redeclare $isStrict */');
        $isStrict = $this->structure->isStrict();
        for ($i=0;$i<count($keys);$i++)
        {
            if ($isStrict === true && !$this->structure->isColumn($this->table, $keys[$i])) {
                trigger_error("SQL syntax error. The column '".$keys[$i]."' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return array();
            } else {
                $insert[strtoupper($keys[$i])] = $values[$i];
            }
        }

        assert('is_array($insert)');
        return $insert;
    }

    /**
     * _increment
     *
     * simulate MySQL's auto-increment feature
     *
     * @access private
     * @param array $res
     * @return array
     * @ignore
     */
    function _increment($res)
    {
        settype($res, "array");
        $primary_key = strtoupper($this->structure->getPrimaryKey($this->table));

        if (is_null($this->auto_increment)) {
         $this->auto_increment = new counter('file_db_auto_increment',YANA_COUNTER_IGNORE_IP);
        }

        /* simulate auto-increment */
        if (empty($res[$primary_key])) {
            $index = $this->auto_increment->count($this->database.'__'.$this->table);
            $res[$primary_key] = $index;
        }

        return $res;
    }

    /**
     * _isset
     *
     * simulate select 1 from ... where ...
     *
     * @access private
     * @param string $tableName
     * @param string $where
     * @return boolean
     * @ignore
     */
    function _isset($tableName, $where)
    {
        assert('is_string($tableName);');
        assert('is_string($where);');
        /* simple case - no further restrictions */
        if (preg_match('/^(\S+) IS NOT NULL$/', $where, $m)) {
            $result = $this->_get($m[1]);
            if (count($result) > 0) {
                return true;
            } else {
                return false;
            }
        /* complex case - restriction by where clause */
        } else if (preg_match('/^(.+?)(?: AND (\S+) IS NOT NULL)?$/', $where, $m)) {
            if (!empty($m[1])) {
                $where = $this->_where($m[1]);
                if ($where === false) {
                    /* syntax error */
                    return false;
                }
            } else {
                $where = array();
            }
            if (empty($m[2])) {
                /**
                 * {@internal
                 *
                 * currently used:
                 *
                 * $what = '*';
                 *
                 * instead of:
                 *
                 * $what = $this->getPrimaryKey($this->table);
                 *
                 * Note: due to the current implementation this single operation (which
                 * takes about 0.0005 to 0.0006 seconds on my current machine) is about
                 * 5 to 7% faster if the whole resultset ('*') is taken, instead of
                 * limiting (filtering) the resultset to a certain column (like the
                 * primary key).
                 *
                 * I mention this here because it is quite the opposite of what one might
                 * expect in the first place.
                 *
                 * The current implementation should be revised, in case the underlaying
                 * implementation has changed to reflect on the performance.
                 *
                 * }}
                 */
                $what = '*';
            } else {
                $what = $m[2];
            }
            $result = $this->_get($what,$where);
            if (count($result) > 0) {
                return true;
            } else {
                return false;
            }
        } else {
            /* syntax error */
            return false;
        }

    }

    /**
     * _length
     *
     * simulate select count(*) from ... where ...
     *
     * @access private
     * @param  string  $tableName
     * @param  array   $where
     * @return int
     * @ignore
     */
    function _length($tableName, $where)
    {
        assert('is_string($tableName);');
        assert('is_array($where);');
        settype($tableName,"string");
        settype($where,    "array");
        
        $tableName = strtolower(trim($tableName));
        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$tableName])) {
            return 0;
        }

        if (empty($where)) {
            $smlfile =& $this->src[$this->database][$tableName];
            return $smlfile->length($tableName);
        } else {
            $result = $this->_get('*',$where);
            return count($result);
        }
    }

    /**
     * _get
     *
     * @access private
     * @param  string $what
     * @param  array  $where
     * @param  int    $offset
     * @param  int    $limit
     * @return array
     * @ignore
     */
     function _get($what = "*", $where = array(), $offset=0, $limit=0)
    {
        assert('is_string($what);');
        assert('is_array($where);');
        assert('is_int($offset) && $offset >= 0;');
        assert('is_int($limit) && $limit >= 0;');
        settype($what,   "string");
        settype($where,  "array");
        settype($limit,  "integer");
        settype($offset, "integer");

        /* if table does not exist, then there is nothing to get */
        if (!isset($this->src[$this->database][$this->table])) {
            return array();
        }

        /* initialize vars */
        $result      =  array();
        $smlfile     =& $this->src[$this->database][$this->table];
        $primary_key =  $this->structure->getPrimaryKey($this->table);
        $data        =  $smlfile->get($primary_key);

        /* if there is data */
        if (is_array($data)) {
            assert('!isset($i); /* cannot redeclare variable $i */');
            assert('!isset($current); /* cannot redeclare variable $current */');

            /* implements a table-scan */
            foreach ($data as $i => $current)
            {
                $current[strtoupper($primary_key)] = $i;

                /* implements where-clause */
                $test = true;
                if (!empty($where)) {
                    assert('!isset($value); /* cannot redeclare variable $value */');
                    foreach ($where as $value)
                    {
                        /* if all required information is provided */
                        if (count($value) === 3) {
                            assert('!isset($currentValue); /* cannot redeclare variable $currentValue */');
                            if (isset($current[strtoupper($value[0])])) {
                                $currentValue = $current[strtoupper($value[0])];
                            } else {
                                $currentValue = null;
                            }
                            switch ($value[2])
                            {
                                case '=':
                                    if ($this->structure->isPrimaryKey($this->table,$value[0]) || $this->structure->isForeignKey($this->table,$value[0])) {
                                        if (strcasecmp($currentValue, $value[1]) !== 0) {
                                            $test = false;
                                        }
                                    } else {
                                        if (strcmp($currentValue, $value[1]) !== 0) {
                                            $test = false;
                                        }
                                    }
                                break;

                                case 'like':
                                    if (!preg_match('/^'.$value[1].'$/i',$currentValue)) {
                                        $test = false;
                                    }
                                break;

                                case '<':
                                    if (!($currentValue < $value[1])) {
                                        $test = false;
                                    }
                                break;

                                case '>':
                                    if (!($currentValue > $value[1])) {
                                        $test = false;
                                    }
                                break;

                                case '!=':
                                    if ($this->structure->isPrimaryKey($this->table,$value[0]) || $this->structure->isForeignKey($this->table,$value[0])) {
                                        if (strcasecmp($currentValue, $value[1]) === 0) {
                                            $test = false;
                                        }
                                    } else {
                                        if (strcmp($currentValue, $value[1]) === 0) {
                                            $test = false;
                                        }
                                    }
                                break;

                                case '<=':
                                    if (!($currentValue <= $value[1])) {
                                        $test = false;
                                    }
                                break;

                                case '>=':
                                    if (!($currentValue >= $value[1])) {
                                        $test = false;
                                    }
                                break;

                                default:
                                    /* intentionally left blank */
                                break;
                            }
                            unset($currentValue);
                        }
                    } /* end foreach */
                    unset($value);
                } /* end if*/

                /* implements the resultset */
                if ($test === true) {
                    if ($what == '*' || empty($what)) {
                        array_push($result, $current);
                    } else if (isset($current[strtoupper($what)])) {
                        array_push($result, array(strtoupper($what) => $current[strtoupper($what)]));
                    }
                }
            } /* end foreach */
            unset($i, $current);
        } else {
            /* result is empty */
            return array();
        }

        /* sort */

        /* If no sort criteria is provided, the array will be returned order in which the entries
         * where inserted. If descending order is requested, the order will be reversed, so the
         * latest entries will come first and the oldest last.
         */
        if ($this->desc === true && empty($this->sort)) {
            $result = array_reverse($result);
        }
        /* If a column has been provided to sort entries by, then the resultset will get sorted
         * by it.
         */
        if (!empty($this->sort)) {
            usort($result, array($this, '_sort'));
        }

        /* limits and offsets */
        if ($limit>0 || $offset>0) {
            if ($limit<=0) {
                $limit = count($result);
            }
            if ($offset<0) {
                $offset = 0;
            }
            $result = array_slice($result, $offset, $limit);
        }
        return $result;
    }

    /**
     * _write
     *
     * @access private
     * @param boolean $commit (optional) default = false
     * @return filedb_result
     * @ignore
     */
    function _write($commit = false)
    {
        $this->cache = array();
        if (!$commit && !$this->auto_commit) {
            /* wait for commit */
            return new filedb_result(array());
        } else {
            /* auto commit */
            $smlfile =& $this->src[$this->database][$this->table];
            if (!$smlfile->write()) {
                /* error */
                return new filedb_result(null, 'unable to save changes');
            } else {
                $this->idx[$this->database][$this->table]->commit();
                /* success */
                return new filedb_result(array());
            }
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $another_object
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same structure file.
     *
     * @access   public
     * @param    object $another_object
     * @return   string
     */
    function equals($another_object)
    {
        if (is_object($another_object) && method_exists($another_object,'getClass') && $another_object->getClass() === $this->getClass()) {
            if (!isset($this->structure) || !isset($another_object->structure)) {
                return isset($this->structure) === isset($another_object->structure);
            } else if ($this->structure->equals($another_object->structure)) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

}

?>