<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * @ignore
 */

/**
 * FileDb_index
 *
 * This implements index files for table indexes.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @since       2.8.6
 * @ignore
 */
class FileDb_index extends Object
{

    /**#@+
     * @ignore
     * @access  private
     */

    /** @var StructureFile */ var $structure = null;
    /** @var string        */ var $table     = "";
    /** @var SML           */ var $data      = null;
    /** @var SML           */ var $src       = null;

    /**#@-*/

    /**
     * constructor
     *
     * @param string $database (optional)
     */
    function FileDb_index(&$structure, &$table, &$data, $filename)
    {
        if (!is_object($structure) || !is_string($table) || !is_object($data) || !is_string($filename)) {
            trigger_error('Invalid database input.',E_USER_ERROR);
            exit;
        } else {
            $this->structure =& $structure;
            $this->table     =& $table;
            $this->data      =& $data;
            $this->src       =  new SML($filename, CASE_UPPER);
            $this->src->init();
            if ($this->src->exists() === false) {
                $this->src->create();
                $this->src->reset();
            }
        }
    }

    /**
     * create an index
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * Note: This will commit any uncommitted data.
     *
     * @access private
     * @param  string    $column
     * @param  array     $update
     * @return bool
     */
    function create($column=null, $update = null)
    {
        assert('is_null($column) || is_string($column);');
        assert('is_null($update) || is_array($update);');

        /* autoscan */
        if (is_null($column)) {
            if (!is_null($update)) {
                trigger_error('No column name provided. Unable to build index', E_USER_WARNING);
                return false;
            } else {
                $columnList = $this->structure->getIndexes($this->table);
                $columnList = mergeArrays($columnList,$this->structure->getUniqueConstraints($this->table));
                if (is_array($columnList) && count($columnList) > 0) {
                    foreach ($columnList as $columnName)
                    {
                        if ($this->create($columnName) !== true) {
                            return false;
                        }
                    } /* end foreach */
                    return true;
                } else {
                    return false;
                }
            }

        /* no scan */
        } else {

            if (!isset($this->structure)) {
                return false;
            } else if (!is_null($update) && (!is_array($update) || count($update)!==2 || !is_scalar($update[0]) || !is_scalar($update[1]))) {
                return false;
            } else if (!$this->structure->isColumn($this->table,$column)) {
                trigger_error("SQL syntax error. The column '".$column."' does not exist in table '".$this->table."'.", E_USER_WARNING);
                return false;
            } else {
                $smlfile    =& $this->src;
                $primaryKey = $this->structure->getPrimaryKey($this->table);
    
                /* no need to index primary key, it is an index by itself */
                if (strcasecmp($primaryKey,$column)===0) {
                    return false;
                }
    
                $column = strtoupper($column);

                /* create / recreate index */
                if (is_null($update) || !$smlfile->exists($column)) {

                    $dataset = $this->data->get($primaryKey);
                    $data    = array();
                    if (!is_null($update)) {
                        $dataset = mergeArrays($dataset, array(strtoupper($update[0]) => array($column => $update[1])));
                    }
                    assert('is_array($dataset);');

                /* update index */
                } else {

                    $dataset = array(strtoupper($update[0]) => array($column => $update[1]));
                    $data    = $smlfile->get($column);
                    if (!is_array($data)) {
                        $data = array();
                    }

                }

                /* process index */
                foreach ($dataset as $value => $row)
                {
                    /* NULL values are to be ignored */
                    if (isset($row[$column])) {
                        $key = strtoupper($row[$column]);
                        if (isset($data[$key])) {
                            if (is_array($data[$key])) {
                                array_push($data,$value);
                            } else {
                                $data[$key] = array($data[$key],$value);
                            }
                        } else {
                            $data[$key] = $value;
                        }
                    }
                } /* end foreach */
                ksort($data);
                $data = array($column=>$data);
                $smlfile->remove($column);

                if (!$smlfile->insert($data)) {
                    return false;
                } else {
                    return true;
                }
            } /* end if */
        } /* end if */

        return false;
    }

    /**
     * get index for some column
     *
     * Returns an associative array on success,
     * returns bool(false) on error.
     *
     * The resulting array uses the following syntax:
     * array (
     *     column_value => primary_key_value
     * )
     *
     * {@internal
     *
     * Exampe of usage:
     * E.g. given the following SQL-statement:
     *  <code>  SELECT * from myTable where indexedRow = "test"  </code>
     *
     * Handle this request as follows:
     * <code>
     * $primaryKeys = $this->get('indexedRow','test');
     *
     * </code>
     * }}
     *
     * @access private
     * @param  string    $column
     * @param  scalar    $value
     * @return array()
     */
    function get($column,$value=null)
    {
        assert('is_string($column);');
        assert('is_null($value) || is_scalar($value);');
        if (!isset($this->structure)) {
            return false;
        } else if (!is_null($value) && !is_scalar($value)) {
            return false;
        } else if (!$this->structure->isColumn($this->table,$column)) {
            trigger_error("SQL syntax error. The column '".$column."' does not exist in table '".$this->table."'.", E_USER_WARNING);
            return false;
        } else if (!$this->structure->hasIndex($this->table,$column) && !$this->structure->isUnique($this->table,$column)) {
            return array();
        } else {
            $smlfile    =& $this->src;
            $primaryKey = $this->structure->getPrimaryKey($this->table);

            if (strcasecmp($primaryKey,$column)===0) {
                return array();
            }

            if (is_null($value)) {
                $index = $smlfile->get($column);
                if (!is_array($index)) {
                    return false;
                } else {
                    return $index;
                }
            } else {
                $value = strtoupper($value);
                $index = $smlfile->get($column.'.'.$value);
                if (!is_scalar($index)) {
                    return false;
                } else {
                    return $index;
                }
            }
        }
    }

    /**
     * write index changes to disk
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * @access private
     * @return bool
     */
    function commit()
    {
        return $this->src->write();
    }

    /**
     * reset index contents
     *
     * Returns bool(true) on success,
     * returns bool(false) on error.
     *
     * @access private
     * @return bool
     */
    function rollback()
    {
        return $this->src->reset();
    }

}

?>