<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * utility Hashtable
 *
 * This is a static utility class to implement certain
 * operations on hashtables (associative arrays).
 *
 * Example for usage of the $key parameter:
 * <pre>
 * Array {
 *     ID1 => Array {
 *         ID2 => 'value'
 *     }
 * }
 * </pre>
 * To get the string 'value' from the hashtable above
 * use $key = 'ID1.ID2'
 *
 * The wildcard '*' may be used to refer to the hashtable
 * as a whole.
 *
 * @access      public
 * @static
 * @package     yana
 * @subpackage  utilities
 */
class Hashtable extends Utility
{
    /**
     * retrieve a value
     *
     * Finds the value identified by $key and returns it.
     * If the value is not found NULL is returned.
     *
     * @access public
     * @static
     * @param array $hash
     * @param string $key
     * @return mixed
     */
    function &get(&$hash, $key)
    {
        assert('is_array($hash);');
        $result = null;

        if (!Hashtable::_checkKey($key)) {
            $result = null;
            return $result;
        }

        if ($key == '*') {
            return $hash;
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $a =& Hashtable::_get($hash, $list_of_keys);
            return $a;
        }
    }

    /**
     * _get
     *
     * @access private
     * @static
     * @param array $hash
     * @param array $list_of_keys
     * @return mixed
     * @ignore
     */
    function &_get(&$hash, &$list_of_keys)
    {
        $key_name = array_shift($list_of_keys);
        if (isset($hash[$key_name]) && !is_null($hash[$key_name])) {
            $result = &$hash[$key_name];
            if (count($list_of_keys)==0) {
                return $result;
            } else {
                $a =& Hashtable::_get($result, $list_of_keys);
                return $a;
            }
        } else {
            $result = null;
            return $result;
        }
    }

    /**
     * set an element by Reference
     *
     * Sets the element identified by $key
     * to $value by passing it's reference.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete hashtable with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access public
     * @static
     * @name   Hashtable::setByReference
     * @param  array  $hash
     * @param  string $key
     * @param  mixed  $value
     * @return boolean
     * @see    Hashtable::set
     * @since  2.8.5
     */
    function setByReference(&$hash, $key, &$value)
    {
        assert('is_array($hash);');

        if (!Hashtable::_checkKey($key)) {
            return false;
        }

        if ($key == '*') {
            if (is_array($value)) {
                $hash = &$value;
                return true;
            } else {
                trigger_error("Wrong argument type: only values of type array may be assigned to a Hashtable. Found '".gettype($value)."' instead.", E_USER_WARNING);
                return false;
            }
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $result = &$hash;
            for ($i=0; $i < count($list_of_keys)-1; $i++)
            {
                if (!isset($result[$list_of_keys[$i]]) || is_scalar($result[$list_of_keys[$i]])) {
                    $result[$list_of_keys[$i]] = array();
                }
                $result = &$result[$list_of_keys[$i]];
            }
            $result[$list_of_keys[count($list_of_keys)-1]] = &$value;
            return true;
        }
    }

    /**
     * set an element to a value
     *
     * Sets the element identified by $key
     * to $value.
     * If the value does not exist it gets inserted.
     * If a previous value existed the value gets updated.
     *
     * This function returns bool(false) if $key = '*'
     * and $value is not an array - which is: trying
     * overwrite the complete hashtable with a non-array value.
     * It returns bool(true) otherwise.
     *
     * @access public
     * @static
     * @name   Hashtable::set
     * @param  array  $hash
     * @param  string $key
     * @param  mixed  $value
     * @return boolean
     * @see    Hashtable::setByReference
     */
    function set(&$hash, $key, $value)
    {
        return Hashtable::setByReference($hash, $key, $value);
    }

    /**
     * set the data type of an element
     *
     * Set the data type of the element identified by $key
     * to $type.
     *
     * Returns bool(false) if the element is NULL or does not exist,
     * or the $type parameter is invalid. Returns bool(true) otherwise.
     *
     * @access public
     * @static
     * @param array $hash
     * @param string $key
     * @param string $type
     * @return boolean
     */
    function setType(&$hash, $key, $type)
    {
        assert('is_string($type) && !empty($type);');
        if (!is_string($type)) {
            trigger_error("Wrong argument type: expect argument 3 '\$type' to be of type string. Found '".gettype($type)."' instead.", E_USER_WARNING);
            return false;
        } else {
            $field =& Hashtable::get($hash, $key);
            if (!is_null($field)) {
                return settype($field, $type);
            } else {
                return false;
            }
        }
    }

    /**
     * check if an element exists
     *
     * Returns bool(false) if the element identified by $key
     * can not be found in the given hashtable or it is NULL.
     * Returns bool(true) otherwise.
     *
     * @access public
     * @static
     * @param array $hash
     * @param string $key
     * @return boolean
     */
    function exists(&$hash, $key)
    {
        return !is_null(Hashtable::get($hash, $key));
    }

    /**
     * remove an element
     *
     * Unsets the element identified by $key in the
     * hashtable. Returns bool(false) if the element
     * does not exist or the key is invalid.
     * Returns bool(true) otherwise.
     *
     * @access public
     * @static
     * @param array $hash
     * @param string $key
     * @return boolean
     */
    function remove(&$hash, $key)
    {
        assert('is_array($hash);');

        if (!Hashtable::_checkKey($key)) {
            return false;
        }

        if ($key == '*') {
            unset($hash);
            return true;
        } else {
            $list_of_keys = explode(".", $key);
            assert('is_array($list_of_keys) && count($list_of_keys) > 0;');
            $stack = array();
            $stack[0] = &$hash;
            $result = &$hash;
            for ($i=0; $i < count($list_of_keys)-1; $i++)
            {
                if (!isset($result[$list_of_keys[$i]])) {
                    return false;
                } else {
                    $result = &$result[$list_of_keys[$i]];
                    $stack[count($stack)] = &$result;
                }
            }
            $j = array_pop($list_of_keys);
            if (!isset($stack[count($stack)-1][$j])) {
                return false;
            } else {
                unset($stack[count($stack)-1][$j]);
                array_pop($stack);
                for ($i=count($stack)-1; $i > -1; $i--)
                {
                    $j = array_pop($list_of_keys);
                    if (is_array($stack[$i][$j]) && count($stack[$i][$j])==0) {
                        unset($stack[$i][$j]);
                    } else {
                        break;
                    }
                }
                return true;
            }
        }
    }

    /**
     * _checkKey
     *
     * @access private
     * @static
     * @param string $key
     * @return boolean
     * @ignore
     */
    function _checkKey($key)
    {
        assert('is_string($key);');
        if (!is_string($key) || ($key != '*' && !preg_match("/^[\w\d-_\.\/ \*]+$/" ,$key))) {
            trigger_error("Syntax error: '".gettype($key)."(".$key.")' is not a valid hash-key.", E_USER_ERROR);
            return false;
        } else {
            return true;
        }
    }
}
?>