<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * abstract InputStream
 *
 * This is an abstract super-class for all filesystem wrappers.
 *
 * @abstract
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class InputStream extends Object
{
    /**#@+
     * @ignore
     */

    /**
     * @access  protected
     * @var     string */  var $filename   = "";
    /**
     * @access  protected
     * @var     array */   var $attributes = array();

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param string $filename
     */
    function InputStream($filename)
    {
        assert('is_string($filename);');
        settype($filename, "string");
        $this->filename = $filename;
        $this->resetStats();
    }

    /**
     * abstract read
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @access public
     * @abstract
     * @return boolean
     *
     * @ignore
     */
    function read()
    {
        /* intentionaly left blank */
        return true;
    }

    /**
     * return stream contents
     *
     * @access public
     * @return string
     */
    function get()
    {
        $text = "IO-Stream \"".$this->filename."\"\n";
        foreach ($this->attributes as $a => $b)
        {
            $text .= "\t- $a = ".gettype($b)."(".$b.")\n";
        }
        return $text;
    }

    /**
     * abstract toString
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @access public
     * @abstract
     * @return string
     *
     * @ignore
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Path to '".$this->filename."' does not exist\n";
        } elseif ($this->isEmpty()) {
            return "Path to '".$this->filename."' exists, but it is wether empty or not loaded\n";
        } else {
            return $this->get();
        }
    }

    /**
     * get filename
     *
     * @access public
     * @return string
     */
    function getFilename()
    {
        assert('is_string($this->filename);');
        return ( $this->filename );
    }

    /**
     * return true, if the input stream resource exists
     *
     * @access public
     * @return boolean
     */
    function exists()
    {
        return file_exists($this->getFilename());
    }

    /**
     * abstract isEmpty
     *
     * Since this is an abstract method, it is to implemented in the sub-classes.
     *
     * @access public
     * @abstract
     * @return boolean
     *
     * @ignore
     */
    function isEmpty()
    {
        return ( false );
    }

    /**
     * reset file statistics
     *
     * Reset file stats, e.g. after creating a file that did not exist.
     * While this method is never used in THIS class, it will propably
     * be essential for its sub classes.
     *
     * @access protected
     *
     * @ignore
     */
    function resetStats()
    {
        if (file_exists($this->filename) && is_readable($this->filename)) {
            clearstatcache();
            $this->attributes['FILESIZE'] = filesize($this->filename);
            $this->attributes['MTIME']    = filemtime($this->filename);
            clearstatcache();
        }
    }


    /**
     * get time when file was last modified
     *
     * Returns the file MTIME value (from cached value).
     * The result is an UNIX timestamp (UTC),
     * or bool(false) if an error occurs.
     *
     * @access public
     * @return int
     * @since  2.8.5
     */
    function getLastModified()
    {
        if (isset($this->attributes['MTIME'])) {
            return $this->attributes['MTIME'];
        } else {
            return false;
        }
    }

    /**
     * compare with another object
     *
     * Returns bool(true) if this object and $another_object
     * are equal and bool(false) otherwise.
     *
     * Two instances are considered equal if and only if
     * they are both objects of the same class and they both
     * refer to the same filesystem resource.
     *
     * @access   public
     * @param    object $another_object
     * @return   string
     */
    function equals($another_object)
    {
        if (is_object($another_object) && method_exists($another_object,'getClass') && $another_object->getClass() === $this->getClass()) {
            return $this->getFilename() === $another_object->getFilename();
        } else {
            return false;
        }
    }

}

?>