<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * abstract Report
 *
 * This is an abstract super class for all kinds of messages.
 *
 * @abstract
 * @access      public
 * @package     yana
 * @subpackage  error_reporting
 */
class Report
{
    /**#@+
     * @access  public
     */

    /**
     * the message that should be reported
     * @var string */ var $message = "";

    /**
     * any kind of data that might help to understand the context in which the message was created
     * @var mixed */  var $data    = null;

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance, representing a system message.
     * To report this message to the framework, just call $YANA->report($report_object).
     *
     * @param string $message the message that should be reported
     * @param mixed $data (optional) any kind of data that might help to understand context in which the message was created
     */
    function Report($message, $data = null)
    {
        assert('is_string($message);');
        settype($message, "string");
        $this->message = $message;
        $this->data    = $data;
    }

    /**
     * create log
     *
     * This creates a log entry that is supposed to be written
     * to a log-file.
     *
     * The result is an associative array or bool(false) on error.
     *
     * @access public
     * @param  string $prefix string to put before each key string
     * @return string
     */
    function getLog($prefix = "")
    {
        /* 0. check input */
        if (!is_string($prefix)) {
            trigger_error("Wrong argument type for argument 1. String expected found '".gettype($prefix)."' instead.", E_USER_WARNING);
            return false;
        }
        /* nothing to report */
        if (empty($this->message) && empty($this->data)) {
            return false;
        }

        /* 1. initialize vars */
        global $YANA;
        $report = array();
        $prefix = strtoupper($prefix);

        /* 2. create entry */
        if (!empty($this->message)) {
            $report[$prefix.'MESSAGE']= $this->message;
        }
        if (!is_null($this->data)) {
            $report[$prefix.'DATA']   = $this->data;
        }
        if (isset($YANA)) {
            $report[$prefix.'USER']   = $YANA->getVar("SESSION_USER_ID");
        }
        if (empty($report[$prefix.'USER'])) {
            $report[$prefix.'USER']   = '*GUEST';
        }
        if (isset($YANA)) {
            $report[$prefix.'ACTION'] = $YANA->getVar("ACTION");
        } else {
            $report[$prefix.'ACTION'] = "*default";
        }
        assert('is_array($report);');
        return $report;
    }

    /**
     * returns the message string of this event
     *
     * This creates a string of at most 1000 characters
     * containing the text in the message attribute and a serialized
     * representation of the data attribute.
     *
     * @access public
     * @return string
     */
    function getMessage()
    {
        $message = $this->message;
        if (!empty($this->data)) {
            if (is_scalar($this->data)) {
                $message .= ";".$this->data;
            } else if (is_array($this->data)) {
                $message .= ";".str_replace("\n", '', SML::encode($this->data));
            } else {
                $message .= ";".print_r($this->data, true);
            }
        }
        $message = str_replace('"', '&quot;',substr(preg_replace('/\s/', ' ', $message),0,1000));
        return $message;
    }

    /**
     * test for equality
     *
     * The parameter $otherLog needs to be an array
     * created by Report::getLog().
     *
     * Checks if the input array and the log created
     * from the current object are the same.
     *
     * The entries are considered identical if and
     * only if: the textual representation of
     * the "message" and "data" values are the
     * same.
     *
     * @access public
     * @param  array  $otherLog
     * @param  string $prefix string to put before each key string
     * @return bool
     * @ignore
     */
    function logEquals($otherLog, $prefix = "")
    {
        /* 0. check input */
        if (!is_array($otherLog)) {
            trigger_error("Wrong argument type for argument 1. Array expected found '".gettype($otherLog)."' instead.", E_USER_NOTICE);
            return false;
        }
        if (!is_string($prefix)) {
            trigger_error("Wrong argument type for argument 2. String expected found '".gettype($prefix)."' instead.", E_USER_NOTICE);
            return false;
        }

        /* 1. initialize vars */
        $prefix = strtoupper($prefix);

        /* 2. return result */
        $thisLog = $this->getLog();
        if (print_r(@$thisLog['MESSAGE'],true) === print_r(@$otherLog[$prefix.'MESSAGE'],true) && strcasecmp(print_r(@$thisLog['DATA'],true), print_r(@$otherLog[$prefix.'DATA'],true)) === 0) {
            return true;
        } else {
            return false;
        }
    }

}

?>