<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * checked reading and writing of file sources
 *
 * This class adds functionality to write changes
 * on the object to the corresponding file.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class SecureFileStream extends SecureInputStream
{

    /**
     * insert new content
     *
     * @access public
     * @param scalar $text
     */
    function insert($text)
    {
        if (is_scalar($text)) {
            if (!isset($this->content)) {
                $this->content = array();
            }
            assert('is_array($this->content);');
            array_push($this->content, $text);
        } else {
            trigger_error("Scalar value expected for parameter 1, found '".gettype($text)."' instead.", E_USER_ERROR);
        }
    }

    /**
     * write file to system
     *
     * This function will return bool(true) on success.
     * It issues an E_USER_NOTICE and returns bool(false) on error.
     *
     * @access public
     * @return boolean
     */
    function write()
    {
        assert('empty($this->content) || is_array($this->content) || is_string($this->content); /* Content attribute has illegal value. Array or string expected. */');
        settype($this->content, "array");

        if (!$this->is_writeable()) {
            trigger_error("Unable to write to file '".$this->getFilename()."'. The file is not writeable.", E_USER_NOTICE);
            return false;
        } else if ($this->getLastModified() != filemtime($this->getFilename())) {
            trigger_error("Unable to write to file '".$this->getFilename()."'.\n\t\tThe file has been changed by some third party recently. Your cached copy is out of date.", E_USER_NOTICE);
            return false;
        } else if ($this->md5Checksum() !== $this->md5Checksum($this->getFilename())) {
            trigger_error("Unable to write to file '".$this->getFilename()."'.\n\t\tThe file on the filesystem has been changed. Your cached copy is whether outdated or invalid.", E_USER_NOTICE);
            return false;
        } else {
            $handle = fopen($this->getFilename(), "w+");
            flock($handle, 2);
            fwrite($handle, implode("", $this->content));
            flock($handle, 3);
            fclose($handle);
            $this->md5 = $this->md5Checksum($this->getFilename());
            if (method_exists($this, 'resetStats')) {
                $this->resetStats();
            }
            return true;
        }
    }

    /**
     * get size of this file
     *
     * Returns the size of the file in bytes (from cached value).
     * If an error occurs, bool(false) is returned.
     *
     * @access public
     * @return int
     * @since  2.8.5     
     */
    function getFilesize()
    {
        if (isset($this->attributes['FILESIZE'])) {
            return $this->attributes['FILESIZE'];
        } else {
            return false;
        }
    }

    /**
     * delete this file
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access public
     * @return boolean
     */
    function delete()
    {
        if (@unlink($this->getFilename())) {
            return true;
        } else {
            trigger_error("Unable to delete file '".$this->getFilename()."'.".((!is_writeable($this->filename))?"\n\t\tThe file is not writeable.":''), E_USER_NOTICE);
            return false;
        }
    }

    /**
     * failSafe writing of data
     *
     * Automatically restarts writing if the file-resource
     * is temporarily not available after waiting for 0.5 seconds.
     *
     * The process is aborted if it failed 3 times.
     *
     * @access public
     * @return boolean
     */
    function failSafeWrite()
    {
        for ($i=0; $i<3; $i++)
        {
            if ($this->write()) {
              return true;
              break;
            } else {
              sleep(0.7);
            }
        }
        return false;
    }

    /**
     * create the current file if it does not exist
     *
     * @access public
     * @return boolean
     */
    function create()
    {
        if (!$this->exists() && !empty($this->filename)) {
            $fhandle = fopen($this->filename, "w+");
            if ($fhandle !== false) {
                fclose($fhandle);
                $this->resetStats();
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * return bool(true) if file is writeable
     *
     * @access public
     * @return boolean
     */
    function is_writeable()
    {
        return ( $this->exists() && is_writeable($this->filename) );
    }

    /**
     * get the number of contents inside the file
     *
     * @access public
     * @return integer
     */
    function length()
    {
        if (!isset($this->content)) {
            return 0;
        } else {
            assert('is_array($this->content);');
            return count($this->content);
        }
    }

    /**
     * remove an entry from the file
     *
     * @access public
     * @param string $key (optional)
     * @return boolean
     */
    function remove($key = "")
    {
        settype($key, "string");

        if ($key == "" || $key == "*") {
            $this->content = array();
            return true;
        } else {
            if (is_numeric($key)) {
                if (!isset($this->content[$key])) {
                    return false;
                } else {
                    array_splice($this->content, $key, 1);
                    return true;
                }
            } else {
                return Hashtable::remove($this->content, $key);
            }
        }

    }

    /**
     * copy the file to some destination
     *
     * This will create a copy of this file on the filesystem.
     * Bool(true) will be returned on success and bool(false)
     * on error.
     *
     * Possible errors are:
     * <ul>
     *   <li>  the destination directory does not exist (and $is_recursive is not provided, or set to false)  </li>
     *   <li>  another file with the same name does already exist (and $overwrite is set to false)  </li>
     *   <li>  the destination file is not accessible for some other reason  </li>
     * </ul>
     *
     * If the directory the desination file would be place in
     * does not exist, and $is_recursive is set to true,
     * it will automatically, recursively create the missing directories.
     * Newly created directories will be set to access restriction 0766.
     * Note that this is unlike the default behaviour, where there is no restriction at all.
     *
     * Note: instead of a file name in parameter $destFile, you may also provide the name of a directory.
     * In this case the file the current name of the file will be used.
     * The destination directory needs to be terminated by a forward-slash '/'.
     *
     * The $destFile parameter must not be any longer than 512 characters.
     * It must not contain any but alphanumeric characters.
     *
     * If you copy a PHP file, note that for security reasons the $mode parameter defaults
     * to 0766, which means the file will not be executable in a UNIX environment (while it may
     * be executable on windows platforms) until you set the $mode parameter to 0777.
     *
     * @access   public
     * @param    string   $destFile      destination to copy the file to
     * @param    bool     $overwrite     setting this to false will prevent existing files from getting overwritten
     * @param    bool     $is_recursive  setting this to true will automatically, recursively create directories in the $destFile string, if required
     * @param    int      $mode          the access restriction that applies to the copied file, defaults to 0766
     * @return   boolean
     *
     * @since    2.8.5
     */
    function copy($destFile,$overwrite=true,$is_recursive=false,$mode=766)
    {
        /* type checking */
        if (!is_string($destFile)) {
            trigger_error("Wrong argument type for argument 1 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($destFile)."' instead.", E_USER_WARNING);
            return false;
        }
        if (!is_bool($overwrite)) {
            trigger_error("Wrong argument type for argument 2 in ".__CLASS__."->".__FUNCTION__."(). Boolean expected, found '".gettype($overwrite)."' instead.", E_USER_WARNING);
            return false;
        }
        if (!is_bool($is_recursive)) {
            trigger_error("Wrong argument type for argument 3 in ".__CLASS__."->".__FUNCTION__."(). Boolean expected, found '".gettype($is_recursive)."' instead.", E_USER_WARNING);
            return false;
        }
        if (!is_int($mode)) {
            trigger_error("Wrong argument type for argument 4 in ".__CLASS__."->".__FUNCTION__."(). Integer expected, found '".gettype($mode)."' instead.", E_USER_WARNING);
            return false;

        }
        /* validity checking */
        if (empty($destFile) || strlen($destFile) > 512 || !preg_match('/^[\w\d-_\.][\w\d-_\/\.]*$/',$destFile)) {
            trigger_error("Invalid filename '".$destFile."' in ".__CLASS__."->".__FUNCTION__."().", E_USER_WARNING);
            return false;
        }
        if (!preg_match('/^[1-7]{1}[0-7]{2}$/',"$mode")) {
            trigger_error("Invalid mode (access level) '".$mode."' in ".__CLASS__."->".__FUNCTION__."().\n\t\tThis argument needs to be an integer of 3 or 4 digits between 0-7.\n\t\tNote: when providing 4 digits, the first digit needs to be 0.", E_USER_WARNING);
            return false;
        }

        $destDir  = dirname($destFile).'/';
        $destFile = preg_replace('/^.*\//s', '', $destFile);
        if (empty($destFile)) {
            $destFile = $this->getFilename();
            $destFile = preg_replace('/^.*\//s', '', $destFile);
        }

        /* check if file already exists */
        if ($overwrite === false && file_exists($destDir.$destFile) === true) {
            trigger_error("Unable to copy to file '${destDir}${destFile}'. Another file with the same name does already exist.",E_USER_NOTICE);
            return false;
        } else if ($overwrite === true && file_exists($destDir.$destFile) === true && is_writeable($destDir.$destFile) === false) {
            trigger_error("Unable to copy to file '${destDir}${destFile}'. The file does already exist and is not writeable.",E_USER_WARNING);
            return false;
        }

        assert('is_string($destDir) && is_string($destFile);');

        /* recursively create directories */
        if (!empty($destDir) && !is_dir($destDir)) {
            if ($is_recursive === true) {
                assert('!isset($currentDir); // cannot redeclare variable $currentDir');
                $currentDir = '';
                assert('!isset($current); // cannot redeclare variable $dir');
                assert('!isset($dir); // cannot redeclare variable $dir');
                foreach (explode('/',$destDir) as $dir)
                {
                    if (!is_dir($currentDir.$dir)) {
                        $current = new DirStream($currentDir.$dir);
                        if ($current->create($mode) === false) {
                            trigger_error("Unable to create directory '".$current->getFilename()."'.",E_USER_NOTICE);
                            return false;
                        }
                    }
                    $currentDir .= $dir.'/';
                } /* end foreach */
                unset($dir,$current);
            } else {
                trigger_error("Unable to copy file '${destFile}'. The directory '${destDir}' does not exist.",E_USER_NOTICE);
                return false;
            }
        }

        if (copy($this->getFilename(),$destDir.$destFile)) {
            if (chmod($destDir.$destFile,"0$mode")===false) {
                trigger_error("Unable to set mode (access level) for file '${destDir}${destFile}'.",E_USER_NOTICE);
            }
            return true;
        } else {
            return false;
        }
    }

}
?>