<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * checked reading of file sources
 *
 * This class adds functionality to compute the checksum
 * of a resource and failsafe reading of resources.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class SecureInputStream extends InputStream
{

    /**#@+
     * @ignore
     * @access  protected
     */

    /** @var string */ var $md5     = null;
    /** @var array  */ var $content = null;
    /** @var array  */ var $options = array();

    /**#@-*/

    /**
     * read file contents
     *
     * Returns bool(true) on success and
     * bool(false) on error.
     *
     * @access public
     * @return boolean
     */
    function read()
    {
        if (!$this->exists()) {
            return false;
        } else {
            $this->content = file($this->filename);
            $this->options = $this->attributes;
            /* calculate hash */
            $this->md5     = $this->md5Checksum();
            return true;
        }
    }

    /**
     * read file contents
     *
     * Automatically restarts reading if the file-resource
     * is temporarily not available after waiting for 0.5 seconds.
     *
     * The process is aborted if it fails 3 times.
     *
     * @access public
     * @return boolean
     */
    function failSafeRead()
    {
        for ($i=0; $i<3; $i++)
        {
            if ($this->read()) {
              return true;
              break;
            } else {
              sleep(0.5);
            }
        }
        return false;
    }

    /**
     * get file contents
     *
     * @access public
     * @return string
     */
    function get()
    {
        if (!$this->isEmpty()) {
            return implode("", $this->content);
        } else {
            return "";
        }
    }

    /**
     * alias of get()
     *
     * @access public
     * @return string
     */
    function toString()
    {
        if (!$this->exists()) {
            return "File ".$this->filename." does not exist\n";
        } elseif ($this->isEmpty()) {
            return "File ".$this->filename." is wether empty or not loaded\n";
        } else {
            return $this->get();
        }
    }

    /**
     * returns bool(true) if the source is empty or not loaded
     *
     * @access public
     * @return boolean
     */
    function isEmpty()
    {
        if (is_null($this->content)) {
            return true;
        } elseif (is_array($this->content)) {
            return count($this->content) == 0;
        } elseif (is_scalar($this->content)) {
            return empty($this->content);
        } else {
            /* the source is invalid */
            return true;
        }
    }

    /**
     * return crc32 checksum for this file
     *
     * The filename parameter became available
     * in version 2.8.5
     *
     * @access public
     * @param  string $filename
     * @return int
     */
    function checksum($filename = "")
    {
        if (!is_string($filename)) {
            trigger_error("Wrong argument type for argument 1 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($filename)."' instead.", E_USER_WARNING);
            return false;
        } else if (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate MD5 hash. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            $filename = $this->getFilename();
        }
        
        $source = implode("",file($filename));

        if (empty($source)) {
            return 0;
        } elseif (is_string($source)) {
            return crc32($source);
        } else {
            /* the source is invalid */
            return 0;
        }
    }

    /**
     * return md5 checksum for this file
     *
     * Usefull to check whether or not a
     * file has been changed since last access.
     *
     * Note: results are cached for the file itself.
     *
     * @access public
     * @return string
     * @since  2.8.5
     */
    function md5Checksum($filename = "")
    {
        /* input checking*/
        if (!is_string($filename)) {
            trigger_error("Wrong argument type for argument 1 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($filename)."' instead.", E_USER_WARNING);
            return false;
        }

        if (!empty($filename)) {
            if (!is_file($filename)) {
                trigger_error("Unable to calculate MD5 hash. The file '${filename}' does not exist.", E_USER_NOTICE);
                return false;
            }
        } else {
            if (isset($this->md5)) {
                assert('is_string($this->md5);');
                return $this->md5;
            } else {
                $filename = $this->getFilename();
                assert('is_file($filename);');
            }
        }
        if (function_exists('md5_file')) {
            $md5 = md5_file($filename);
        } else {
            $md5 = md5(implode("",file($filename)));
        }
        
        if (!is_string($md5)) {
            return false;
        } else {
            return $md5;
        }
    }

    /**
     * reload file contents
     *
     * You accidently did something wrong with the file?
     * Calling this will reload the file from disk while
     * reseting its current state.
     *
     * @access public
     */
    function reset()
    {
        $this->read();
    }

}

?>