<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

if (!defined('CASE_MIXED')) {
    /**
     * @ignore
     */
    define('CASE_MIXED', -1);
}

/**
 * Simple Markup Language (SML) Files
 *
 * This is a wrapper-class that may be used to work with *.config and *.sml files.
 *
 * SML files provide the same semantics and functionality as JSON encoded files,
 * are as easy to read and understand, but they stick with XML-style markup,
 * which is widely used and understood by most people.
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 * @since       2.8.5
 * @name        SML
 */
class SML extends SecureFileStream
{
    /**#@+
     * @access  private
     * @ignore
     */

    /** @var bool  */ var $is_ready       = false;
    /** @var array */ var $value          = array();
    /** @var int   */ var $case_sensitive = CASE_MIXED;

    /**#@-*/

    /**
     * constructor
     *
     * Create a new instance of this class.
     * This extends the super class constructor.
     *
     * Note the additional parameter $case_sensitive.
     * This parameter decides on how to return key names.
     *
     * It's value can be one of the following constants:
     * <ul>
     *     <li>  CASE_UPPER  upper-case all keys       </li>
     *     <li>  CASE_LOWER  lower-case all keys       </li>
     *     <li>  CASE_MIXED  leave keys in mixed case  </li>
     * </ul>
     *
     * @param  string $filename
     * @param  int    $case_sensitive
     */
    function SML($filename, $case_sensitive = CASE_MIXED)
    {
        /* call superclass constructor */
        $constructor = get_parent_class(__CLASS__);
        parent::$constructor($filename);

        if (!is_int($case_sensitive)) {
            trigger_error("Wrong argument type for argument 2. String expected found '".gettype($key)."' instead.", E_USER_WARNING);
            return false;
        }
        if ($case_sensitive === CASE_MIXED || $case_sensitive === CASE_UPPER || $case_sensitive === CASE_LOWER) {
            $this->case_sensitive = $case_sensitive;
        } else {
            trigger_error("Invalid argument 1. Expected one of CASE_MIXED, CASE_UPPER, CASE_LOWER, found '".print_r($case_sensitive,true)."' instead.", E_USER_WARNING);
        }
    }

    /**
     * get a value from the file
     *
     * Returns the value at the position specified by $key.
     * If key is the character '*' or empty, than the whole
     * document is returned.
     *
     * Example: use "foo1.foo2" to get the contents of the "foo2"
     * child tag inside the "foo1" root tag.
     * Or, more technical speaking, to get the "foo2" element
     * of the "foo1" array.
     *
     * @access  public
     * @param   string $key (optional)
     * @return  mixed
     *
     * @name    SML::get()
     */
    function get($key = "*")
    {
        /* input checking */
        $key = $this->_checkKey($key);
        if ($key === false) {
            return false;
        }

        /* auto-load */
        if (!$this->is_ready) {
            $this->init();
        }

        /* return result */
        if ($key === "*") {
            return $this->value;
        } else {
            return Hashtable::get($this->value, $key);
        }
    }

    /**
     * insert an array into the file
     *
     * This merges the current file content with the
     * provided array. New values will replace old ones.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   array  $array
     * @return  boolean
     *
     * @name    SML::insert()
     */
    function insert($array)
    {
        /* input checking */
        $array = $this->_checkArray($array);
        if ($array === false) {
            return false;
        }

        /* return result */
        $value = mergeArrays($this->value,$array);
        $content = SML::encode($value,null,CASE_MIXED);
        if (is_array($value) && is_string($content)) {
            $this->value = $value;
            $this->content = $content;
            return true;
        } else {
            return false;
        }
    }

    /**
     * set the content of the file
     *
     * Replaces the file content with the provided array.
     *
     * Returns bool(true) on success and bool(false) on error.
     *
     * @access  public
     * @param   array  $array
     * @return  boolean
     *
     * @name    SML::set()
     */
    function set($array)
    {
        /* input checking */
        $array = $this->_checkArray($array);
        if ($array === false) {
            return false;
        }

        /* return result */
        $content = SML::encode($array,null,CASE_MIXED);
        if (is_string($content)) {
            $this->value = $array;
            $this->content = $content;
            return true;
        } else {
            return false;
        }
    }

    /**
     * reset the file
     *
     * Changes to the file will not be safed unless you
     * explicitely call $configFile->write().
     * So if you want or need to revert your changes just call
     * $configFile->reset() and all will be fine.
     *
     * @access  public
     *
     * @name    SML::reset()
     */
    function reset()
    {
        /* reset $this->content */
        switch ($this->case_sensitive)
        {
            case CASE_UPPER: case CASE_LOWER: case CASE_MIXED:
                $this->content = SML::encode($this->value,null,$this->case_sensitive);
            break;
            default:
                trigger_error("The program encountered an internal error. Invalid object attribute value in ".__CLASS__."->case_sensitive.", E_USER_ERROR);
                $this->content = "";
            break;
        }
        /* reset $this->options */
        $this->options = $this->attributes;
    }

    /**
     * get a string representation
     *
     * @access  public
     * @return  string
     *
     * @name    SML::toString()
     */
    function toString()
    {
        if (!$this->exists()) {
            return "Config-file ".$this->filename." does not exist";
        } else if ($this->isEmpty()) {
            return "Config-file ".$this->filename." is wether empty or not loaded";
        } else {
            switch ($this->case_sensitive)
            {
                case CASE_UPPER: case CASE_LOWER: case CASE_MIXED:
                    return SML::encode($this->value,null,$this->case_sensitive);
                break;
                default:
                    trigger_error("The program encountered an internal error. Invalid object attribute value in ".__CLASS__."->case_sensitive.", E_USER_ERROR);
                    return "Config-file ".$this->filename." is broken";
                break;
            }            
        }
    }

    /**
     * initialize file contents
     *
     * You should always call this before anything else.
     *
     * @access  public
     * @return  mixed
     *
     * @name    SML::init()
     */
    function init()
    {
        /* auto-load */
        if ($this->isEmpty()) {
            $this->read();
        }
        if (!$this->exists()) {
            return false;
        }

        /* setting $this->value */
        switch ($this->case_sensitive)
        {
            case CASE_UPPER: case CASE_LOWER: case CASE_MIXED:
                $this->value = SML::getFile($this->getFilename(),$this->case_sensitive);
            break;
            default:
                trigger_error("The program encountered an internal error. Invalid object attribute value in ".__CLASS__."->case_sensitive.", E_USER_ERROR);
                $this->value = array();
            break;
        }

        /* setting $this->ready state */
        $this->is_ready = true;
        
        /* return result */
        return $this->value;

    }

    /**
     * get the number of elements
     *
     * This returns how many elements can
     * be found inside the array at position
     * $key. If $key points to a non-existing value,
     * or an empty array, the function returns 0.
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  int
     *
     * @name    SML::length()
     */
    function length($key = "*")
    {
        /* input checking */
        $key = $this->_checkKey($key);
        if ($key === false) {
            return false;
        }

        /* return result */
        if ($key === "*") {
            return count($this->value);
        } else {
            $result =& Hashtable::get($this->value, $key);
            if (is_null($result)) {
                return 0;
            } else {
                assert('is_array($result);');
                return count($result);
            }
        }
    }

    /**
     * remove an entry from the file
     *
     * @access  public
     * @param   string  $key  (optional)
     * @return  boolean
     *
     * @name    SML::remove()
     */
    function remove($key = "*")
    {
        /* input checking */
        $key = $this->_checkKey($key);
        if ($key === false) {
            return false;
        }

        /* compute request */
        if (Hashtable::remove($this->value, $key)) {
            $this->content = SML::encode($this->value,null,CASE_MIXED);
            return true;
        } else {
            return false;
        }

    }

    /**
     * test if a certain value exists
     *
     * This function has two synopsis:
     *
     * 1st: if parameter $key is provided.
     *
     * <ul>
     *   <li>  returns bool(true) if the file exists, is loaded and the variable identified by $key is set  </li>
     *   <li>  returns bool(false) otherwise.  </li>
     * </ul>
     *
     * 2nd: if parameter $key is missing, == '' or the wildcard '*'.
     *
     * <ul>
     *   <li>  returns bool(true) if the file exists and is loaded  </li>
     *   <li>  returns bool(false) otherwise.  </li>
     * </ul>
     *
     * @access  public
     * @param   string $key (optional)
     * @return  boolean
     *
     * @name    SML::exists()
     */
    function exists($key = '*')
    {
        /* input checking */
        $key = $this->_checkKey($key);
        if ($key === false) {
            return false;
        }

        /* return result */
        if ($key === '*') {
            return parent::exists();
        } else {
            return Hashtable::exists($this->value, $key);
        }

    }

    /**
     * Read a file in SML syntax and return its contents
     *
     * The argument $input can wether be a filename or a numeric array
     * of strings created by file($filename).
     *
     * The argument $case_sensitive can be used to decide how keys should be treated.
     *
     * Valid values for $case_sensitive are:
     * <ul>
     *     <li>  CASE_UPPER  upper-case all keys       </li>
     *     <li>  CASE_LOWER  lower-case all keys       </li>
     *     <li>  CASE_MIXED  leave keys in mixed case  </li>
     * </ul>
     *
     * @access  public
     * @static
     * @name    SML::getFile()
     * @param   array|string  $input               filename or file content
     * @param   int           $case_sensitive      CASE_UPPER|CASE_LOWER|CASE_MIXED
     * @return  array
     *
     * @see     SML::encode()
     */
    function getFile($input, $case_sensitive = CASE_MIXED)
    {
        assert('$case_sensitive === CASE_MIXED || $case_sensitive === CASE_LOWER || $case_sensitive === CASE_UPPER;');
    
        /* $translated_key holds the debugging backtrace */
        if (defined('YANA_ERROR_REPORTING') && YANA_ERROR_REPORTING !== YANA_ERROR_OFF) {
            $translated_key = array();
            $i = 0;
            $is_valid = true;
        }
        $result       = array();
        $stack        = array(&$result);
        $stack_length = 0;
        $match1       = array();
        $match2       = array();
        $handle       = null;
        $buffer       = "";
        $is_file      = false;
    
        if (is_file("$input")) {
            $handle  = fopen("$input", "r");;
            $is_file = true;
        } else if (is_array($input)) {
            $is_file = false;
            $i = 0;
        } else {
            trigger_error("Argument 1 is expected to be a filename or an array created with file(). Instead found ".gettype($input)." '".print_r($input,true)."'.", E_USER_ERROR);
            exit;
        }
    
        while (true)
        {
            if ($is_file === true) {
                if (feof($handle)) {
                    break;
                }
                $buffer = fgets($handle);
            } else {
                if (!isset($input[$i])) {
                    break;
                }
                $buffer = &$input[$i];
            }
            if (preg_match("/^\s*<([^\/][^>]*)>/", $buffer, $match1)) {
                /* START TAG */
                if ($case_sensitive === CASE_UPPER) {
                    $match1[1] = strtoupper($match1[1]);
                } else if ($case_sensitive === CASE_LOWER) {
                    $match1[1] = strtolower($match1[1]);
                } else {
                    /* intentionally left blank */
                }
                $stack[$stack_length +1] =& $stack[$stack_length++][$match1[1]];
    
                if (preg_match('/^'.preg_quote($match1[0]).'(.*)<\/'.preg_quote($match1[1]).'>/Ui', $buffer, $match2)) {
                    /* CDATA */
                    if ($match2[1] == "true") {
                        $stack[$stack_length] = true;
                    } elseif ($match2[1] == "false") {
                        $stack[$stack_length] = false;
                    } elseif ($match2[1]!="") {
                        $stack[$stack_length] = $match2[1];
                    }
                    array_pop($stack);
                    $stack_length--;
                } else {
                    /* debugging backtrace */
                    if (defined('YANA_ERROR_REPORTING') && YANA_ERROR_REPORTING !== YANA_ERROR_OFF) {
                        array_push($translated_key, $match1[1]);
                    }                
                }
    
            } else if (preg_match("/<\/[^\/]*>/U", $buffer)) {
                /* END TAG */
                array_pop($stack);
                $stack_length--;
                /* debugging backtrace */
                if (defined('YANA_ERROR_REPORTING') && YANA_ERROR_REPORTING !== YANA_ERROR_OFF) {
                    $open_tag = array_pop($translated_key);
                    /* hide follow up errors */
                    if ($is_valid === true) {
                        assert('!isset($m);');
                        preg_match("/<\/([^\/]*)>/U", $buffer, $m);
                        $closed_tag = $m[1];
                        unset($m);
                        if (strcasecmp($open_tag, $closed_tag) !== 0) {
                            if ($is_file) {
                                trigger_error("SML ERROR in file '".$input."' on line ".$i.": Unclosed tag '".implode('.',$translated_key).'.'.$open_tag."'.", E_USER_WARNING);
                            } else {
                                trigger_error("SML ERROR on line ".$i.": Unclosed tag '".implode('.',$translated_key).'.'.$open_tag."'.", E_USER_WARNING);
                            }
                            $is_valid = false;
                        }
                    }
                }
            } else {
                /* COMMENT */
            }
            if ($is_file === false || (defined('YANA_ERROR_REPORTING') && YANA_ERROR_REPORTING !== YANA_ERROR_OFF)) {
                $i++;
            }
        }

        if ($is_file) {
            fclose($handle);
        }
    
        if (defined('YANA_ERROR_REPORTING') && YANA_ERROR_REPORTING !== YANA_ERROR_OFF && $stack_length !== 0) {
            if ($is_file) {
                trigger_error("SML ERROR in file '${input}': The tag '".implode('.',$translated_key)."' has never been closed.", E_USER_WARNING);
            } else {
                trigger_error("SML ERROR: The tag '".implode('.',$translated_key)."' has never been closed.", E_USER_WARNING);
            }
        }
    
        return $result;
    
    }

    /**
     * Create a SML string from a scalar variable or an array of data.
     *
     * The argument $name can be used to specify the name of the root node.
     * If $name is omitted, no root node is created.
     *
     * Note that this function will issue an E_USER_NOTICE if $name is omitted
     * and $data is a scalar value. In this case the scalar variable will
     * be named '0' by default.
     *
     * The argument $case_sensitive can be used to decide how keys should be treated.
     *
     * Note that any tags from string inputs will be stripped.
     * You should convert tags to entities, before submiting the input.
     *
     * Valid values for $case_sensitive are:
     * <ul>
     *     <li>  CASE_UPPER  upper-case all keys       </li>
     *     <li>  CASE_LOWER  lower-case all keys       </li>
     *     <li>  CASE_MIXED  leave keys in mixed case  </li>
     * </ul>
     *
     * @access  public
     * @static
     * @name    SML::encode()
     * @param   scalar|array  $data
     * @param   string        $name
     * @param   int           $case_sensitive
     *
     * @see     SML::getFile()
     */
    function encode($data, $name = null, $case_sensitive = CASE_MIXED, $indent = 0)
    {
        assert('is_null($data) || is_scalar($data) || is_array($data);');
        assert('$case_sensitive === CASE_MIXED || $case_sensitive === CASE_LOWER || $case_sensitive === CASE_UPPER;');
    
        settype($name,   "string");
        settype($indent, "integer");
        settype($case_sensitive,   "int");
    
        /* string */ $txt = "";
    
        if (is_null($data)) {
            return "";
        }
        if (is_scalar($data) && (is_null($name) || $name === "")) {
            trigger_error("Your untitled scalar value (arg. 1) in function ".__FUNCTION__."() was renamed to '0'.\n\t\tYou are encouraged to use the \$name argument (arg. 2) to set the variable name to anything you prefer.",E_USER_NOTICE);
            $name = '0';
        }
    
        if ($case_sensitive === CASE_UPPER) {
            $name = strtoupper($name);
        } else if ($case_sensitive === CASE_LOWER) {
            $name = strtolower($name);
        }
    
        if ((is_null($name) || $name === "") && is_array($data)) {
            $ignore_name = true;
        } else {
            $ignore_name = false;
        }
    
        /* indent tag */
        $tab = "";
        for ($i=0;$i<$indent;$i++)
        {
            $tab .= "\t";
        }
    
        if (is_bool($data)) {
            return $tab."<$name>".( ($data) ? "true" : "false" )."</$name>"."\n";
        } else if (is_string($data)) {
            return $tab."<$name>".preg_replace("/\s/", " ", strip_tags($data))."</$name>"."\n";
        } else if (is_integer($data) || is_float($data)) {
            return $tab."<$name>".$data."</$name>"."\n";
        } else if (is_array($data)) {
            $ignore_name || $txt = $tab."<$name>"."\n";
            $ignore_name || $indent++;
            foreach ($data as $key=>$element)
            {
                if ($element !== $data) {
                    $txt .= SML::encode($element, $key, $case_sensitive, $indent);
                } else {
                    /* ignore recursion */
                }
            } /* end foreach */
            $ignore_name || $txt .= $tab."</$name>"."\n";
            return $txt;
        } else if (is_object($data) && method_exists ($data, "serialize")) {
            return $tab."<$name>".$data->serialize()."</$name>"."\n";
        } else {
            return $tab."<$name>".$data."</$name>"."\n";
        }
    }

    /**
     * Read variables from an encoded string
     *
     * This function is pretty much the same as SML::getFile() except
     * for the fact that it is working on strings rather than files.
     *
     * Returns NULL on error.
     *
     * The argument $input has to be a string, that has been encoded using
     * SML::encode().
     *
     * The argument $case_sensitive can be used to decide how keys should be treated.
     *
     * Valid values for $case_sensitive are:
     * <ul>
     *     <li>  CASE_UPPER  upper-case all keys       </li>
     *     <li>  CASE_LOWER  lower-case all keys       </li>
     *     <li>  CASE_MIXED  leave keys in mixed case  </li>
     * </ul>
     *
     * Note: to reaccess an encoded value look at the following examples.
     *
     * Handling boolean values:
     * <code>
     * $input_bool = true;
     * $encoded = SML::encode($input_bool, 'MY_VAR');
     * $decoded = SML::decode($encoded);
     * // the following returns true
     * $input_bool === $decoded['MY_VAR'];
     * </code>
     *
     * ... or shorter:
     * <code>
     * $input_bool = true
     * // the following returns true
     * $input_bool === array_pop(SML::decode(SML::encode($input_bool)));
     * </code>
     *
     * Handling string values and nummerics:
     * <code>
     * $input_string = 'foo';
     * // the following returns true
     * $input_string === array_pop(SML::decode(SML::encode($input_string)));
     *
     * $input_int = 123;
     * // the following returns true
     * $input_int == array_pop(SML::decode(SML::encode($input_int)));
     * </code>
     *
     * Handling the 'NULL' value:
     * <code>
     * $input_null = null;
     * // the following returns true
     * is_null( array_pop(SML::decode(SML::encode($input_string))) );
     * </code>
     *
     * Arrays (were key case does matter):
     * <code>
     * $input_array = array(1,2,3,array(4,5),'a'=>6,'B'=>7);
     * $output_array = SML::decode(SML::encode($input_array));
     * // the following returns true
     * $input_array == $output_array;
     * </code>
     *
     * When dealing with nummeric arrays, or associative arrays where all keys should be uppercase,
     * or if you just don't care, you may set the $case_sensitive parameter to CASE_UPPER.
     *
     * <code>
     * $input_array = array(1,2,3,array(4,5),'A'=>6,'B'=>7);
     * $output_array = SML::decode(SML::encode($input_array,null,CASE_UPPER),CASE_UPPER);
     * // the following returns true
     * $input_array == $output_array;
     * </code>
     *
     * The obvious advantage of doing so is: you can rely on the writing of keys with no need to care
     * for case-sensitivity.
     *
     * @access  public
     * @static
     * @name    SML::decode()
     * @param   string    $input
     * @param   int       $case_sensitive
     * @return  array
     *
     * @see     SML::getFile()
     * @see     SML::encode()
     */
    function decode($input, $case_sensitive = CASE_MIXED)
    {
        if (!is_string($input)) {
            trigger_error("Wrong argument type for argument 1. String expected, found '".gettype($input)."' instead.", E_USER_WARNING);
            return null;
        } else {
            $input = explode("\n", $input);
            return SML::getFile($input,$case_sensitive);
        }
    }

    /**
     * validate and convert $key parameter
     *
     * <ul>
     *   <li>  returns modified $key as string   </li>
     *   <li>  returns bool(false) if key is not valid  </li>
     * </ul>
     *
     * @access  private
     * @param   string  $key
     * @return  bool(false)|string
     * @ignore
     */
    function _checkKey($key)
    {
        /* input checking*/
        if (!is_string($key)) {
            trigger_error("Wrong argument type for argument 1. String expected found '".gettype($key)."' instead.", E_USER_WARNING);
            return false;
        }

        /* Convert empty strings to wildcard.
         * This is to evade problems when testing for bool(false)
         * E.g. "" == false; will compute to true.
         */
        if (empty($key) || $key === '*') {
            return '*';
        }

        /* create and return output */
        switch ($this->case_sensitive)
        {
            case CASE_UPPER:
                return strtoupper($key);
            break;
            case CASE_LOWER:
                return strtolower($key);
            break;
            default:
                return $key;
            break;
        }

    }

    /**
     * validate and convert $array parameter
     *
     * <ul>
     *   <li>  returns modified $array as array   </li>
     *   <li>  returns bool(false) if array is not valid  </li>
     * </ul>
     *
     * @access  private
     * @param   array   $array
     * @return  bool(false)|array
     * @ignore
     */
    function _checkArray($array)
    {
        /* input checking*/
        if (!is_array($array)) {
            trigger_error("Wrong argument type for argument 1. Array expected found '".gettype($array)."' instead.", E_USER_WARNING);
            return false;
        }

        /* create and return output */
        switch ($this->case_sensitive)
        {
            case CASE_UPPER:
                return changeCaseArray($array,CASE_UPPER);
            break;
            case CASE_LOWER:
                return changeCaseArray($array,CASE_LOWER);
            break;
            default:
                return $array;
            break;
        }

    }

}

$lcn_note = urldecode('%3Cp+class%3D%22KOMMENTAR%22+style%3D%22text-align%3Acenter%22%3Eregistered+VIP-version%3A+%3Cspan+class%3D%22BESCHRIFTUNG%22%3E%26quot%3B%25OWNER%25%26quot%3B%3C%2Fspan%3E%3C%2Fp%3E');
?>