<?php
/**
 * YANA library
 *
 * @package yana
 * @license http://creativecommons.org/licenses/by-nc-sa/2.5/
 */

/**
 * database schema file
 *
 * This wrapper class represents the structure of a database
 *
 * @access      public
 * @package     yana
 * @subpackage  io
 */
class StructureFile extends SML
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var  array  */   var $cached_fields = array();

    /**#@-*/


    /**
     * constructor
     *
     * Create a new instance of this class.
     *
     * @param string $filename
     */
    function StructureFile($filename)
    {
        /* call superclass constructor */
        $constructor = get_parent_class(__CLASS__);
        parent::$constructor($filename, CASE_UPPER);

        /* specific extensions */
        assert('is_string($filename);');
        $this->filename = StructureFile::_getFilename($filename);
        $this->resetStats();
    }


    /**
     * initialize the file
     *
     * Always call this first before anything else.
     *
     * @access public
     * @return boolean
     */
    function init()
    {
        if (parent::init() !== false) {
            foreach ($this->value['TABLES'] as $k => $v)
            {
                $this->value['TABLES'][$k]['PRIMARY_KEY'] = strtolower($v['PRIMARY_KEY']);
                if (isset($v['FOREIGN_KEYS'])) {
                    assert('!isset($k1); /* cannot redeclare variable $k1 */');
                    assert('!isset($v1); /* cannot redeclare variable $v1 */');
                    foreach ($v['FOREIGN_KEYS'] as $k1 => $v1)
                    {
                        $this->value['TABLES'][$k]['FOREIGN_KEYS'][$k1] = strtolower($v1);
                    }
                    unset($k1,$v1);
                }
            }
            if (isset($this->value['INCLUDE'])) {
                if (is_string($this->value['INCLUDE'])) {
                    $test = $this->addFile($this->value['INCLUDE']);
                    if ($test === false) {
                        trigger_error("Error importing file '".$this->value['INCLUDE']."'.", E_USER_WARNING);
                    }
                } else if (is_array($this->value['INCLUDE'])) {
                    foreach ($this->value['INCLUDE'] as $filename)
                    {
                        if (is_string($filename)) {
                            $test = $this->addFile($filename);
                            if ($test === false) {
                                trigger_error("Error importing file '${filename}'.", E_USER_WARNING);
                            }
                        } else {
                            trigger_error("Invalid field type in ".__CLASS__."->".__FUNCTION__."().\n\t\tContents of array 'INCLUDE' are supposed to be strings. Found '".gettype($filename)."' instead.", E_USER_NOTICE);
                        }
                    } /* end foreach */
                } else {
                    trigger_error("Invalid field type in ".__CLASS__."->".__FUNCTION__."().\n\t\tField 'INCLUDE' is supposed to be array or string. Found '".gettype($filename)."' instead.", E_USER_NOTICE);
                }
            } /* end if*/
            return $this->value;
        } else {
            return false;
        }
    }

    /**
     * get the compiled structure of the database
     *
     * @access public
     * @return mixed
     */
    function getStructure()
    {
        return $this->value;
    }

    /**
     * get the file source
     *
     * Returns the text of the source file containing
     * the database structure as a string or bool(false)
     * on error.
     *
     * @access public
     * @return mixed
     */
    function getSource()
    {
        $source = file($this->getFilename());
        if (!is_array($source)) {
            return false;
        } else {
            $source = implode("",$source);
            assert('is_string($source);');
            if (empty($source)) {
                return false;
            } else {
                return $source;
            }
        }
    }

    /**
     * check whether a table exists in the current structure
     *
     * Returns bool(true) if a table with the given name is listed
     * in the current structure file and bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param string $table
     * @return boolean
     */
    function isTable($table)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        return isset($this->value['TABLES'][$table]);
    }

    /**
     * check whether a column exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * in its list of contents. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return boolean
     */
    function isColumn($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        return isset($this->value['TABLES'][$table]['CONTENT'][$column]);
    }

    /**
     * check whether a column is indexed in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column,
     * which has an index. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return boolean
     * @since  2.8.5
     */
    function hasIndex($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        if (isset($this->value['TABLES'][$table]['CONTENT'][$column]['INDEX'])) {
            if (is_string($this->value['TABLES'][$table]['CONTENT'][$column]['INDEX'])) {
                if (strcasecmp($this->value['TABLES'][$table]['CONTENT'][$column]['INDEX'], 'true') === 0) {
                    return true;
                } else {
                    return false;
                }
            } else if ($this->value['TABLES'][$table]['CONTENT'][$column]['INDEX'] == true) {
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }
    }

    /**
     * check whether a foreign key exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and it has a column named $column
     * in its list of foreign keys. Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return boolean
     */
    function isForeignKey($table, $column)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        settype($column, 'string');
        $column = strtoupper($column);
        return isset($this->value['TABLES'][$table]['FOREIGN_KEYS'][$column]);
    }

    /**
     * check whether a primary key exists in the current structure
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file and its primary key is named $column.
     * Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param  string $table
     * @param  string $column
     * @return boolean
     */
    function isPrimaryKey($table, $column)
    {
        settype($column, 'string');
        return (strcasecmp($this->getPrimaryKey($table), $column) === 0);
    }

    /**
     * check whether a column has a unique constraint
     *
     * Returns bool(true) if a table named $table is listed
     * in the current structure file, it has a column named $column
     * in its list of contents and this column has a unique constraint.
     * Returns bool(false) otherwise.
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param  string $table
     * @param  string $column
     * @return boolean
     * @since  2.8.5
     */
    function isUnique($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        return isset($this->value['TABLES'][$table]['CONTENT'][$column]['UNIQUE']) && ($this->value['TABLES'][$table]['CONTENT'][$column]['UNIQUE'] === true || $this->value['TABLES'][$table]['CONTENT'][$column]['UNIQUE'] === 'true');
    }

    /**
     * return a list of foreign keys defined on a table
     *
     * Returns an associative array of foreign keys.
     * If $table is not a specified in the current structure file,
     * then NULL is returned instead.
     * If the table has no foreign keys, an empty array is returned.
     *
     * Note that this operation is not case sensitive.
     *
     * @access public
     * @param string $table
     * @return array
     */
    function getForeignKeys($table)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        if (!isset($this->value['TABLES'][$table])) {
            trigger_error("Unable to get foreign keys. There is no table '${table}' in current structure file.", E_USER_WARNING);
            return null;
        } else if (isset($this->value['TABLES'][$table]['FOREIGN_KEYS'])) {
            return $this->value['TABLES'][$table]['FOREIGN_KEYS'];
        } else {
            return array();
        }
    }

    /**
     * get the primary key of a table
     *
     * Returns the name of the primary key column of $table as a
     * lower-cased string.
     * Returns NULL and issues an E_USER_WARNING if $table is not a listed table in the current
     * structure file.
     * Returns NULL and issues an E_USER_WARNING if there is no primary key for $table.
     *
     * @access public
     * @param string $table
     * @return string
     */
    function getPrimaryKey($table)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        if (!isset($this->value['TABLES'][$table])) {
            trigger_error("Unable to get primary key. There is no table '${table}' in current structure file.", E_USER_WARNING);
            return null;
        } else if (!isset($this->value['TABLES'][$table]['PRIMARY_KEY'])) {
            trigger_error("Table '${table}' has no primary key declaration.", E_USER_WARNING);
            return null;
        } else {
            return strtolower($this->value['TABLES'][$table]['PRIMARY_KEY']);
        }
    }

    /**
     * get the maximum length of a field as specified in the structure
     *
     * Returns the 'length' attribute of $column in $table.
     * Returns int(0) if there is no $table, or the table has no
     * column named $column, or the column does not have a 'length'
     * attribute.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return int
     */
    function getLength($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        if (isset($this->value['TABLES'][$table]['CONTENT'][$column]['LENGTH'])) {
            $length = $this->value['TABLES'][$table]['CONTENT'][$column]['LENGTH'];
            assert('is_numeric($length);');
            settype($length, 'integer');
        } else {
            $length = 0;
        }
        return $length;
    }

    /**
     * get the data type of a field as specified in the structure
     *
     * Returns the 'type' attribute of $column in $table as a lower-cased string.
     * Returns string("") if there is no $table, or the table has no
     * column named $column, or the column does not have a 'type'
     * attribute.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return string
     */
   function getType($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        if (isset($this->value['TABLES'][$table]['CONTENT'][$column]['TYPE'])) {
            return strtolower($this->value['TABLES'][$table]['CONTENT'][$column]['TYPE']);
        } else {
            return '';
        }
    }

    /**
     * get the default value of a field as specified in the structure
     *
     * Returns constant NULL (not bool(false)) if value is not specified.
     * Use is_null($result) to test for existance, don't use empty($result).
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return mixed
     */
   function getDefault($table, $column)
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        if (isset($this->value['TABLES'][$table]['CONTENT'][$column]['DEFAULT'])) {
            return $this->value['TABLES'][$table]['CONTENT'][$column]['DEFAULT'];
        } else {
            return null;
        }
    }

    /**
     * get the names of all columns in a table
     *
     * Returns a numeric array of all columns in $table.
     * Issues an E_USER_NOTICE and returns an empty array, if $table does not exist
     * in current structure file.
     *
     * @access public
     * @param string $table
     * @return array
     */
   function getColumns($table)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        if (!$this->isTable($table)) {
            /* table does not exist */
            trigger_error("The table '${table}' does not exist.",E_USER_NOTICE);
            return array();
        } else {
            return array_keys(array_change_key_case($this->value['TABLES'][$table]['CONTENT'],CASE_LOWER));
        }
    }

    /**
     * get the name of the table, a foreign key points to
     *
     * @access public
     * @param string $table
     * @param string $foreignKey
     * @return string
     */
    function getTableByForeignKey($table, $foreignKey)
    {
        settype($table, 'string');
        settype($foreignKey, 'string');
        $table = strtoupper($table);
        $foreignKey = strtoupper($foreignKey);
        return strtolower(@$this->value['TABLES'][$table]['FOREIGN_KEYS'][$foreignKey]);
    }

    /**
     * check whether the structure defines the "USE_STRICT" setting as bool(true)
     *
     * @access public
     * @return boolean
     */
    function isStrict()
    {
        return ($this->value['USE_STRICT'] === true || $this->value['USE_STRICT'] === 'true');
    }

    /**
     * get a list of all tables in the current database
     *
     * Returns a numeric array of all tables in the current structure file.
     * Issues an E_USER_NOTICE and returns an empty array, if the list of
     * tables is empty.
     *
     * @access public
     * @return array
     */
    function getTables()
    {
        if (!isset($this->value['TABLES'])) {
            trigger_error("The list of tables is empty.",E_USER_NOTICE);
            return array();
        } else {
            return array_keys(array_change_key_case($this->value['TABLES'],CASE_LOWER));
        }
    }

    /**
     * get a list of all indexed columns in a table
     *
     * Returns a numeric array of all columns in $table, that are marked with the
     * option INDEX => true (which means, all columns that have an index).
     *
     * @access public
     * @param  string  $table
     * @return array
     * @since  2.8.5
     */
    function getIndexes($table)
    {
        if (!isset($this->value['TABLES'])) {
            trigger_error("The list of tables is empty.",E_USER_NOTICE);
            return array();
        } else if (!$this->isTable($table)) {
            /* table does not exist */
            trigger_error("The table '${table}' does not exist.",E_USER_NOTICE);
            return array();
        } else {
            $result = array();
            $table = strtoupper($table);
            foreach ($this->getColumns($table) as $columnName)
            {
                if ($this->hasIndex($table,$columnName) === true) {
                    array_push($result,$columnName);
                }
            } /* end foreach */
            return $result;
        }
    }

    /**
     * get a list of all unique columns in table
     *
     * Returns a numeric array of all columns in $table, that are marked with the
     * option UNQIUE => true (which means, all columns that have an unique constraint).
     *
     * @access public
     * @param  string  $table
     * @return array
     * @since  2.8.5
     */
    function getUniqueConstraints($table)
    {
        if (!isset($this->value['TABLES'])) {
            trigger_error("The list of tables is empty.",E_USER_NOTICE);
            return array();
        } else if (!$this->isTable($table)) {
            /* table does not exist */
            trigger_error("The table '${table}' does not exist.",E_USER_NOTICE);
            return array();
        } else {
            $result = array();
            $table = strtoupper($table);
            foreach ($this->getColumns($table) as $columnName)
            {
                if ($this->isUnique($table,$columnName) === true) {
                    array_push($result,$columnName);
                }
            } /* end foreach */
            return $result;
        }
    }

    /**
     * get the database name associated with a table
     *
     * returns the associated database name as string or null of no association is found
     *
     * @access protected
     * @return string
     * @ignore
     */
    function getAssociation($table)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        if (!isset($this->value['ASSOCIATIONS']) || !is_array($this->value['ASSOCIATIONS']) || !isset($this->value['ASSOCIATIONS'][$table]) || !is_string($this->value['ASSOCIATIONS'][$table])) {
            return null;
        } else {
            return $this->value['ASSOCIATIONS'][$table];
        }
    }

    /**
     * get all constraints for an address
     *
     * Retrieves all "constraint" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * @access public
     * @since  version 2.8.4
     * @param  string $operation one of select, insert, update, delete
     * @param  string $table
     * @param  array  $columns (optional)
     * @return array
     */
    function getConstraint($operation, $table, $columns = array())
    {
        return $this->_getFields('constraint', $operation, $table, $columns);
    }

    /**
     * get all triggers for an address
     *
     * Retrieves all "trigger" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * There are two constants to use for argument $prefix:
     * The first, YANA_TRIGGER_BEFORE, refers to triggers that fire BEFORE the action
     * $operation is carried out.
     * And YANA_TRIGGER_AFTER, refers to triggers that fire AFTER the action
     * $operation has been carried out.
     *
     * @access public
     * @since  version 2.8.5
     * @param  int    $prefix currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param  string $operation one of insert, update, delete
     * @param  string $table
     * @param  array  $columns (optional)
     * @return array
     */
    function getTrigger($prefix, $operation, $table, $columns = array())
    {
        if (!is_string($operation) || !preg_match('/^select|insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        } else {
            switch ($prefix)
            {
                case YANA_TRIGGER_BEFORE:
                    $operation = 'BEFORE_'.$operation;
                break;
                case YANA_TRIGGER_AFTER:
                    $operation = 'AFTER_'.$operation;
                break;
                default:
                    trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected an integer of value '".YANA_TRIGGER_BEFORE."', or '".YANA_TRIGGER_AFTER."' found (".gettype($prefix).")'".print_r($prefix, true)."' instead.",E_USER_WARNING);
                    return false;
                break;
            }
            return $this->_getFields('trigger', $operation, $table, $columns);
        }
    }

    /**
     * check whether the "READONLY" flag is set to bool(true)
     *
     * Returns bool(true) if:
     * <ol>
     *    <li>database is set to 'readonly=true', or</li>
     *    <li>$table is set to 'readonly=true', or</li>
     *    <li>$column in $table is set to 'readonly=true'</li>
     * </ol>
     *
     * Returns bool(false) otherwise.
     *
     * @access public
     * @param string $table
     * @param string $column
     * @return boolean
     */
    function isReadonly($table="",$column="")
    {
        settype($table, 'string');
        settype($column, 'string');
        $table = strtoupper($table);
        $column = strtoupper($column);
        if (!empty($this->value['READONLY'])) {
            return true;
        } else if (!empty($table) && !empty($column) && !empty($this->value['TABLES'][$table]['CONTENT'][$column]['READONLY'])) {
            return true;
        } else if (!empty($table) && !empty($this->value['TABLES'][$table]['READONLY'])) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * validate a row against the schema
     *
     * Returns bool(true) if $row is valid and
     * bool(false) otherwise.
     *
     * @access public
     * @param string $table
     * @param string &$row
     * @return boolean
     */
    function checkRow($table,&$row, $is_insert = true)
    {
        settype($table, 'string');
        $table = strtoupper($table);
        foreach ($this->value['TABLES'][$table]['CONTENT'] as $column => $b)
        {
            $column = strtolower($column);
            if (!empty($b['REQUIRED']) && (($is_insert && !isset($row[$column])) || (isset($row[$column]) && (is_null($row[$column]) || $row[$column] === "")))) {
                if (strcasecmp($b['REQUIRED'], 'auto') == 0) {
                    switch ($this->getType($table,$column)) {
                        case 'time':
                            $row[$column] = time();
                        break;
                        case 'ip':
                            if (isset($_SERVER['REMOTE_ADDR'])) {
                                $row[$column] = $_SERVER['REMOTE_ADDR'];
                            } else {
                                $row[$column] = '0.0.0.0';
                            }
                        break;
                        default:
                            $row[$column] = false;
                        break;
                    }
                } else if (isset($b['DEFAULT'])) {
                    if (is_array($b['DEFAULT'])) {
                        $row[$column] = $this->untaintInput($table,$column, array_shift(array_keys($b['DEFAULT'])), YANA_ESCAPE_NONE);
                    } else if (is_scalar($b['DEFAULT'])) {
                        $row[$column] = $this->untaintInput($table,$column, $b['DEFAULT'], YANA_ESCAPE_NONE);
                    } else {
                        trigger_error("SYNTAX ERROR: Invalid default value for column '${column}'. The structure file is not valid.", E_USER_ERROR);
                    }
                    if ($b['TYPE'] == 'string') {
                        $row[$column] = Layout::replaceToken($row[$column]);
                    }
                } else {
                    trigger_error("SQL ERROR: A required attribute '${column}' has not been provided. Insert/Update operation aborted.", E_USER_NOTICE);
                    return false;
                } /* end if */
            } else {
                /* this input is valid - move to next */
            } /* end if */
        } /* end for */
        return true;
    }

    /**
     * constraint
     *
     * @since  version 2.8.4
     * @access public
     * @param  string $operation
     * @param  string $table
     * @param  string $field (optional)
     * @param  mixed $value (optional)
     * @return boolean
     * @ignore
     */
    function constraint($operation, $table, $field = NULL, $value = NULL)
    {
        if (!is_string($operation) || !preg_match('/^select|insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'select', 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        }
        return $this->_eval('constraint', $operation, $table, $field, $value);
    }

    /**
     * trigger
     *
     * @since  version 2.8.5
     * @access public
     * @param  int    $prefix currently one of YANA_TRIGGER_BEFORE, YANA_TRIGGER_AFTER
     * @param  string $operation
     * @param  string $table
     * @param  string $field
     * @param  mixed  $value
     * @ignore
     */
    function trigger($prefix, $operation, $table, $field, &$value)
    {
        if (!is_int($prefix)) {
            trigger_error("Wrong argument type for argument 1 in ".__CLASS__."->".__FUNCTION__."(). Integer expected, found '".gettype($prefix)."' instead.",E_USER_WARNING);
            return false;
        }
        if (!is_string($operation) || !preg_match('/^insert|update|delete$/is', $operation)) {
            trigger_error("Invalid input for argument 2 in ".__CLASS__."->".__FUNCTION__."().\n\t\tExpected a string value 'insert', 'update' or 'delete' found (".gettype($operation).")'".print_r($operation, true)."' instead.",E_USER_WARNING);
            return false;
        }
        switch ($prefix)
        {
            case YANA_TRIGGER_BEFORE:
                $operation = 'BEFORE_'. $operation;
            break;
            case YANA_TRIGGER_AFTER:
                $operation = 'AFTER_'. $operation;
            break;
            default:
                trigger_error("Invalid input for argument 1 in ".__CLASS__."->".__FUNCTION__."(). Expected an integer of value '".YANA_TRIGGER_BEFORE."', or '".YANA_TRIGGER_AFTER."' found  ".gettype($prefix)." '".print_r($prefix, true)."' instead.",E_USER_WARNING);
                return false;
            break;
        }

        return $this->_eval('trigger', $operation, $table, $field, $value);

    }

    /**
     * get all fields for an address
     *
     * Retrieves all "$fieldname" entries that apply to the given
     * operation on the dataset and returns the results as an numeric
     * array.
     *
     * If the optional parameter $assoc is set to true, the result
     * will be an associative array.
     *
     * @access private
     * @param  string $fieldname e.g. 'constraint', or 'trigger'
     * @param  string $operation one of select, insert, update, delete
     * @param  string $table
     * @param  array  $columns (optional)
     * @param  bool   $as_assoc (optional)
     * @return array
     * @ignore
     */
    function _getFields($fieldname, $operation, $table, $columns = array(), $as_assoc = false)
    {
        settype($fieldname, "string");
        settype($operation, "string");
        settype($table,     "string");
        settype($columns,   "array");
        settype($as_assoc,  "boolean");

        $fieldname = strtoupper($fieldname);
        $operation = strtoupper($operation);
        $table = strtoupper($table);

        $cache =& $this->cached_fields[(int)$as_assoc][$fieldname][$operation][$table];

        /* if this result has already been cached than there is nothing to do here */
        if (isset($cache)) {
            if (count($columns) === 0) {
                return $cache;
            } else {
                if (!is_array($cache) || count($cache) === 0) {
                    return array();
                }
                $list_of_results = array();
                if (isset($cache[0])) {
                    $list_of_results[0] =& $cache[0];
                }
                if (isset($cache[1])) {
                    $list_of_results[1] =& $cache[1];
                }
                $columns = changeCaseArray($columns, CASE_UPPER);;
                $match = array_intersect(array_keys($cache),$columns);

                foreach ($match as $key => $definition)
                {
                    if ($as_assoc) {
                        $list_of_results[$key] =& $cache[$definition];
                    } else {
                        $list_of_results[count($list_of_results)] =& $cache[$definition];
                    }
                }
                return $list_of_results;
            }
        /* if no cached result is available it has to be resolved */
        } else {

            /* 1. places where fields are allowed */
            $db_table_col = array();

            /* 1.1 Database-wide definitions */
            $db_table_col[0] =& $this->value;

            /* 1.2 table-wide definitions */
            if (!$this->isTable($table)) {
                /* table does not exist */
                trigger_error("The table '${table}' does not exist.",E_USER_NOTICE);
                return array();
            } else {
                $db_table_col[1] =& $this->value['TABLES'][$table];
            }

            /* 1.3 column specific definitions */
            if (!is_array($columns) || count($columns) === 0) {
                $columns = array_keys($this->value['TABLES'][$table]['CONTENT']);
            }

            assert('is_array($columns); // the table '.$table.' has no contents');
            if (is_array($columns)) {
                foreach ($columns as $column)
                {
                    $column = strtoupper($column);
                    if (!$this->isColumn($table, $column)) {
                        /* column does not exist */
                        trigger_error("The column '${column}' does not exist in table '${table}.'",E_USER_NOTICE);
                        return array();
                    } else {
                        $db_table_col[$column] =& $this->value['TABLES'][$table]['CONTENT'][$column];
                    }
                } /* end foreach */
            } /* end if */

            /* results are put here */
            $list_of_results = array();

            /* 2. find definitions */
            foreach ($db_table_col as $key => $value)
            {
                if (isset($value[$fieldname])) {
                    $definition =& $value[$fieldname];
                    /* fieldname.operation = string */
                    if (is_array($definition) && isset($definition[$operation]) && is_scalar($definition[$operation])) {
                        if ($as_assoc) {
                            $list_of_results[$key] =& $definition[$operation];
                        } else {
                            $list_of_results[count($list_of_results)] =& $definition[$operation];
                        }
                    /* fieldname = string */
                    } else if (is_scalar($definition)) {
                        if ($as_assoc) {
                            $list_of_results[$key] =& $definition;
                        } else {
                            $list_of_results[count($list_of_results)] =& $definition;
                        }
                    /* none */
                    } else {
                        /* no definition present */
                    }
                }
            } /* end foreach */

            $cache = $list_of_results;

            return $list_of_results;
        }

    }

    /**
     * evaluate triggers and constraints
     *
     * Use $fieldname parameter to seperate
     * between a constraint check and triggers.
     *
     * @access private
     * @param string $fieldname
     * @param string $operation
     * @param string $table
     * @param string $field
     * @param mixed $value
     * @return boolean
     * @ignore
     */
    function _eval($fieldname, $operation, $table, $field, &$value)
    {
        global $YANA;

        if (!is_string($fieldname)) {
            trigger_error("Wrong argument type for argument 1 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($fieldname)."' instead.",E_USER_WARNING);
            return false;
        } else {
            $fieldname = strtoupper($fieldname);
        }
        if ($fieldname !== 'CONSTRAINT' && $fieldname !== 'TRIGGER') {
            trigger_error("Invalid argument 1 in ".__CLASS__."->".__FUNCTION__."(). String 'CONSTRAINT' or 'TRIGGER' expected, found '".print_r($fieldname,true)."' instead.",E_USER_WARNING);
            return false;
        }
        if (!is_string($operation)) {
            trigger_error("Wrong argument type for argument 2 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($operation)."' instead.",E_USER_WARNING);
            return false;
        } else {
            $operation = strtoupper($operation);
        }
        if (!is_string($table)) {
            trigger_error("Wrong argument type for argument 3 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($table)."' instead.",E_USER_WARNING);
            return false;
        } else {
            $table = strtoupper($table);
        }
        if (!is_null($field) && !is_string($field)) {
            trigger_error("Wrong argument type for argument 4 in ".__CLASS__."->".__FUNCTION__."(). String expected, found '".gettype($field)."' instead.",E_USER_WARNING);
            return false;
        } else {
            settype($field, "string");
            $field = strtoupper($field);
        }

        /* get list of fields */
        if (empty($field)) {
            $list_of_fields = $this->_getFields($fieldname, $operation, $table, array(),true);
        } else {
            $list_of_fields = $this->_getFields($fieldname, $operation, $table, array($field),true);
        }
        assert('is_array($list_of_fields);');

        /* 1.1 prepare value */
        if (is_string($value)) {
            $value = stripslashes($value);
        } else if (is_array($value) && (is_null($field) || empty($field))) {
            $value = changeCaseArray($value, CASE_UPPER);
        }

        /* 1.2 prepare permission */
        if (isset($YANA)) {
            $arg_permission = $YANA->getVar("PERMISSION");
        } else {
            $arg_permission = 0;
        }

        /* 2 evaluate element */
        $test = true;
        foreach ($list_of_fields as $column_name => $element)
        {
            if (!is_scalar($element)) {
                continue;
            } else {
                if ($fieldname == 'CONSTRAINT') {
                    if (is_bool($element)) {
                        $function = create_function('$OPERATION,$PERMISSION,$TABLE,$FIELD,$VALUE', "return ".(($element)?'true':'false').";");
                    } else if (trim($element) == '') {
                        $function = create_function('$OPERATION,$PERMISSION,$TABLE,$FIELD,$VALUE', 'return false;');
                    } else {
                        $element = " $element ";
                        $element = preg_replace("/m(\/.*?\/\S*)/", "preg_match(\"\\1\", \$VALUE)", $element);
                        if (!preg_match("/^((-| |\!)?\\\$[\w\d_]+ | true | false | null | -?\d+ | | \&\&? | (preg_match|ereg|eregi)\(('[^']*'|\"[^\"]*\"), ?\\\$[\w\d_]+\) |[\&\|\!\~\-\*\/\%\+\<\>]| \"[^\"]*\" | \'[^\']*\' | \d+(\.\d*)? |(\=|\!|\<|\>)\=\=?| )*$/i", $element)) {
                            trigger_error("ERROR: in structure file '".$this->filename."', syntax error in constraint: '".trim($element)."'.", E_USER_WARNING);
                            return false;
                        }                        
                        $function = create_function('$OPERATION,$PERMISSION,$TABLE,$FIELD,$VALUE', "if ($element) return true; else return false;");
                    }
                } else {
                    if (is_string($element)) {
                        $function = create_function('$OPERATION,$PERMISSION,$TABLE,$FIELD,&$VALUE', $element);
                    } else {
                        $function = create_function('$OPERATION,$PERMISSION,$TABLE,$FIELD,&$VALUE', 'return true;');
                    }
                }
            }
            /* handling row element */
            if (empty($field) && is_string($column_name)) {
                $column_name = strtoupper($column_name);
                if (!isset($value[$column_name])) {
                    $value[$column_name] = null;
                }
                $col_value =& $value[$column_name];
                switch ($operation)
                {
                    case 'SELECT': case 'DELETE':
                        $test = $function($operation, $arg_permission, $table, $column_name, $col_value);
                    break;
                    case 'INSERT': case 'UPDATE':
                        if (!is_null($col_value)) {
                            $test = $function($operation, $arg_permission, $table, $column_name, $col_value);
                        }
                    break;
                    default:
                    break;
                }
            /* handling database, table and field element */
            } else {
                $test = $function($operation, $arg_permission, $table, $field, $value);
            }
            if (is_bool($test) && $test === false) {
                if (isset($YANA)) {
                    $YANA->report( new log("DB WARNING: Operation '${operation}' stopped by '${fieldname}' on table '${table}'".((!is_null($field))?" at field '${field}'":"").((!is_null($value))?" with value '".print_r($value,true)."'":"")."."));
                    return false;
                }
            }
        }

        if ($test) {
            return true;
        } else {
            return false;
        }
    }

    /**
     * untaint user input data with the help of the schema
     *
     * @access public
     * @param string $table
     * @param string $column
     * @param mixed $value
     * @param int $escape
     * @return mixed
     */
    function untaintInput($table, $column, $value, $escape = YANA_ESCAPE_NONE)
    {
        return untaintInput($value, (string) $this->getType($table,$column), (int) $this->getLength($table,$column), $escape);
    }

    /**
     * include structure file
     *
     * merges the contents of another file with the current structure
     *
     * @access public
     * @param string $filename
     * @return bool
     */
    function addFile($filename)
    {
        $filename = StructureFile::_getFilename($filename);
        if (!is_file($filename) || !preg_match('/^(?:[\w\d\_\-\/]*\/)*([\w\d\_\-]+)\.config$/', $filename, $m)) {
            trigger_error("Illegal structure file '${filename}'.", E_USER_ERROR);
            return false;
        } else {
            assert('!isset($databaseName); // cannot redeclare $databaseName');
            assert('!isset($file); // cannot redeclare $file');
            assert('!isset($test); // cannot redeclare $test');
            $databaseName = $m[1];
            assert('is_string($databaseName) && !empty($databaseName); // Invalid database name');
            $file = new StructureFile($filename);
            $test = $file->init();
            if ($test === false) {
                return false;
            } else {
                assert('!isset($current_list); // cannot redeclare $current_list');
                assert('!isset($import_list); // cannot redeclare $import_list');
                assert('!isset($intersection); // cannot redeclare $intersection');
                $current_list = array_keys($this->value['TABLES']);
                $import_list  = array_keys($file->value['TABLES']);
                $intersection = array_intersect($current_list,$import_list);
                if (is_array($intersection) && count($intersection) > 0) {
                    trigger_error("Unable to import structure file '${filename}' into structure file '".$this->getFilename()."'.\n\t\tAmbigious table declarations found for: ".implode(', ',$intersection), E_USER_WARNING);
                    return false;
                } else {
                    assert('!isset($table); // cannot redeclare $table');
                    if (!isset($this->value['ASSOCIATIONS'])) {
                        $this->value['ASSOCIATIONS'] = array();
                    }
                    foreach ($import_list as $table)
                    {
                        if (isset($this->value['ASSOCIATIONS'][$table])) {
                            trigger_error("For table '${table}' replacing illegal table association '".$this->value['ASSOCIATIONS'][$table]."' with '${databaseName}'.",E_USER_NOTICE);
                        }
                        $this->value['ASSOCIATIONS'][$table] = $databaseName;
                        $this->value['TABLES'][$table] = $file->value['TABLES'][$table];
                    }
                    return true;
                } /* end if */
            } /* end if */
        } /* end if */
    }


    /**
     * addStructure
     *
     * Try to extract some information on the structure of
     * a database from the information provided by PEAR-DB's
     * tableInfo() function.
     *
     * see bufferedDbStream::buildStructure() for additional
     * information on the input
     *
     * @uses void buildStructure()
     *
     * @access public
     * @param string $table current table name
     * @param array $array value returned by getTableInfo() -function
     * @ignore
     */
    function addStructure($table, $array)
    {
        assert('is_array($array);');
        if (!is_array($array)) {
            return false;
        } else {
            $array = changeCaseArray($array,CASE_UPPER);
            $config   = array("USE_STRICT" => "true", "READONLY" => "false", "TABLES" => array());
            for ($i = 0; $i < $array['NUM_FIELDS']; $i++)
            {
               $name  = $array[$i]['NAME'];
               $flags = explode(" ", $array[$i]['FLAGS']);
               $config['TABLES'][$table]['CONTENT'][$name] = array();
               switch ($array[$i]['TYPE']) {
                   case 'blob':
                       $array[$i]['TYPE']                          = "text";
                   break;
                   case 'tinyint': case 'bigint': case 'int':
                       $array[$i]['TYPE']                          = "integer";
                   break;
                   case 'bool':
                       $array[$i]['TYPE']                          = "boolean";
                   break;
                   case 'double': case 'numeric':
                       $array[$i]['TYPE']                          = "float";
                   break;
                   case 'timestamp':
                       $array[$i]['TYPE']                          = "time";
                       $array[$i]['REQUIRED']                      = "AUTO";
                   break;
                   case 'datetime': case 'string': case 'varchar':
                       $array[$i]['TYPE']                          = "string";
                   break;
                   /* more ? */
                   default:
                       /* convert unrecognized type to string */
                       $array[$i]['TYPE']                          = "string";
                   break;

               }
               if (!empty($array[$i]['TYPE'])) {
                   $config['TABLES'][$table]['CONTENT'][$name]['TYPE'] = $array[$i]['TYPE'];
               }
               if (!empty($array[$i]['LEN'])) {
                   $config['TABLES'][$table]['CONTENT'][$name]['LENGTH']= $array[$i]['LEN'];
               }
               $config['TABLES'][$table]['CONTENT'][$name]['REQUIRED'] = in_array('not_null', $flags);
               if (in_array('primary_key', $flags)) {
                    $config['TABLES'][$table]['PRIMARY_KEY']             = $name;
                    $config['TABLES'][$table]['CONTENT'][$name]['HIDDEN']= "true";
               }
            }
            if (isset($this->value)) {
                $this->value = mergeArrays($this->value, $config);
            } else {
                $this->value = $config;
            }
        }
    }

    /**
     * _getFilename
     *
     * check input $filename
     *
     * @static
     * @access private
     * @param string $filename
     * @return string
     * @ignore
     */
    function _getFilename($filename)
    {
        settype($filename, "string");
        if (preg_match('/^([\w\d_]+)$/', $filename)) {
            if (isset($YANA)) {
                $filename = $YANA->getVar('DBDIR').$filename.".config";
            } else {
                $filename = "config/db/".$filename.".config";
            }
        }
        return $filename;
    }

}

?>
