<?php
/**
 * @author     Thomas Meyer
 * @license    http://creativecommons.org/licenses/by-nc-sa/2.5/
 * @package    yana
 * @subpackage plugins
 */

/**
 * RSS to HTML factory plugin
 *
 * creates HTML from RSS files
 *
 * @access     public
 * @package    yana
 * @subpackage plugins
 */
class plugin_rss extends Plugin
{
    /**#@+
     * @ignore
     * @access  private
     */

    /** @var array  */ var $rss;

    /** "file" should be a valid path to an existing XML-File
     * @var string  */ var $file;

    /**  maximum number of entries to show
     * @var int     */ var $max;

    /** @var string */ var $current_tag = "";
    /** @var int    */ var $current_entry = -1;

    /**#@-*/

    /**
     * Constructor
     *
     * @param string $name a name for this plugin
     * @access public
     * @ignore
     */
    function plugin_rss($name)
    {
        settype($name,"string");
        $this->name = $name;
        
        global $YANA;
        $this->file = $YANA->getVar("PROFILE.RSS.FILE");
        $this->max  = $YANA->getVar("PROFILE.RSS.MAX");
        settype($this->file, "string");
        settype($this->max, "integer");
        
        if (!$this->max>0 ) {
            $this->max  = 5;
        }
        if (!$this->file>0) {
            $this->file = 'plugins/rss/test.rss';
        }
    
    }

    /**
     * Default event handler
     *
     * returns bool(true) on success and bool(false) on error
     *
     * @access  public
     * @return  bool
     * @param   string  $event  name of the called event in lower-case
     * @param   array   $ARGS   array of arguments passed to the function
     * @ignore
     */
    function _default($event, $ARGS)
    {
        global $YANA;
        /* do nothing */
        return true;
    }

    /**
     * Transform RSS/XML to HTML
     *
     * This creates the output.
     *
     * This function does not expect any arguments
     *
     * @access public
     * @param array $ARGS array of params passed to the function
     * @return boolean
     */
    function get_news($ARGS)
    {
        /* this function expects no arguments */

        global $YANA;
        $xml_parser = xml_parser_create();
        xml_set_element_handler($xml_parser, array(&$this,"_startElement"), array(&$this,"_endElement"));
        xml_set_character_data_handler($xml_parser, array(&$this,"_characterData"));

        if (!($fp = @fopen($this->file, "r"))) {

            trigger_error("Could not open XML input. Check if '".print_r($this->file)."' is a correct path to a valid XML file!", E_USER_WARNING);
            return false;

        } else {

            while ($data = fread($fp, 4096))
            {
                if (!xml_parse($xml_parser, $data, feof($fp))) {
                    trigger_error("XML error: %s at line %d", xml_error_string(xml_get_error_code($xml_parser)), xml_get_current_line_number($xml_parser), E_USER_WARNING);
                    return false;
                }
            }
            xml_parser_free($xml_parser);
            $YANA->setVar("RSS", array_slice($this->rss,0,$this->max));
            $YANA->setVar("FILE", $this->file);

            /* Microsummaries */
            if (class_exists('Microsummary')) {
                Microsummary::publish($this->name);
            }

            if (class_exists('Microsummary') && count($this->rss) > 0) {
                $latest = array_shift($this->rss);
                Microsummary::set($this->name, strtoupper($this->name).' latest: '.$latest['TITLE'].' ('.$latest['PUBDATE'].')');
            }

            return true;

        } /* end if */
    }

    /**
     * _startElement
     *
     * @access private
     * @param  int    $parser
     * @param  string $name
     * @param  array  $attr
     * @return boolean
     * @ignore
     */
    function _startElement($parser, $name, $attrs)
    {
        if (preg_match("/title|pubDate|description|link|author|category|comments/i", $name)) {
           $this->current_tag = strtoupper($name);
        }
        if (strtoupper($name) == "ITEM" && $this->current_entry == -1) {
            $this->current_entry = 0;
        }
    }

    /**
     * _endElement
     *
     * @access private
     * @param  int    $parser
     * @param  string $name
     * @return boolean
     * @ignore
     */
    function _endElement($parser, $name)
    {
        if (strtoupper($name) == "ITEM") {
            $this->current_entry++;
        }
    }

    /**
     * _characterData
     *
     * @access private
     * @param  int    $parser
     * @param  string $data
     * @return boolean
     * @ignore
     */
    function _characterData($parser, $data)
    {

        if ( trim($data) && $this->current_entry > -1) {
            if ($this->current_tag != "LINK") {
                $data = preg_replace("/'/", "\\'", utf8_decode($data));
            }
            settype($this->rss[$this->current_entry][$this->current_tag], "string");
            $this->rss[$this->current_entry][$this->current_tag] .= $data;
        }
    }

}

//  2006, Thomas Meyer www.All-Community.de
// Copyright Hinweis nicht entfernen!
?>